/* FILE: basedriver.cc            -*-Mode: c++-*-
 *
 * Example concrete Oxs_Driver class.
 *
 */

#include "nb.h"
#include "basedriver.h"
#include "director.h"
#include "simstate.h"
#include "depkey.h"
#include "evolver.h"
#include "mesh.h"
#include "fixedscalarfield.h"
#include "fixedvectorfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_BaseDriver);

/* End includes */

#define BUF_SIZE 1024
static char buf[BUF_SIZE];	// Buffer for building interpreter results

// Constructor
Oxs_BaseDriver::Oxs_BaseDriver(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Driver(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  CheckInitValueParamCount("max_timestep",1);
  max_timestep=Nb_Atof((*FindInitValue("max_timestep"))[0].c_str());
  if(max_timestep<=0.0) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName())
      + ": Specified max time step is"
      + (*FindInitValue("max_timestep"))[0];
    throw Oxs_Ext::Error(msg.c_str());
  }
  DeleteInitValue("max_timestep");

  CheckInitValueParamCount("min_timestep",1);
  min_timestep=Nb_Atof((*FindInitValue("min_timestep"))[0].c_str());
  DeleteInitValue("min_timestep");
  if(min_timestep==0.0) min_timestep=max_timestep*1e-6;

  CheckInitValueParamCount("stopping_dm_dt",1);
  stopping_dm_dt=Nb_Atof((*FindInitValue("stopping_dm_dt"))[0].c_str());
  DeleteInitValue("stopping_dm_dt");


  const vector<string>*  vsptr; // Temp pointer

  number_of_stages=1;  // Default value: 1 stage
  vsptr = FindInitValue("number_of_stages");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("number_of_stages",1);
    number_of_stages=atoi((*vsptr)[0].c_str());
    DeleteInitValue("number_of_stages");
  }

  stage_iteration_limit=0;  // Default value: no limit
  vsptr = FindInitValue("stage_iteration_limit");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("stage_iteration_limit",1);
    stage_iteration_limit=atoi((*vsptr)[0].c_str());
    DeleteInitValue("stage_iteration_limit");
  }

  total_iteration_limit=0;  // Default value: no limit
  vsptr = FindInitValue("total_iteration_limit");
  if(vsptr!=NULL) {
    CheckInitValueParamCount("total_iteration_limit",1);
    total_iteration_limit=atoi((*vsptr)[0].c_str());
    DeleteInitValue("total_iteration_limit");
  }
 
  CheckInitValueParamCount("mesh",1);
  Oxs_DepKey<Oxs_Ext> key1
    = director->FindExtObject((*FindInitValue("mesh"))[0]);
  Oxs_Mesh* mesh = dynamic_cast<Oxs_Mesh*>(key1.GetPtr());
  if(mesh==NULL) {
    string msg="No Oxs_Mesh object found with name "
      + (*FindInitValue("mesh"))[0];
    throw Oxs_Ext::Error(msg.c_str());
  }
  DeleteInitValue("mesh");

  CheckInitValueParamCount("Ms",1);
  Oxs_DepKey<Oxs_Ext> key2
    = director->FindExtObject((*FindInitValue("Ms"))[0]);
  Oxs_FixedScalarField* Ms
    = dynamic_cast<Oxs_FixedScalarField*>(key2.GetPtr());
  if(Ms==NULL) {
    string msg="No Oxs_FixedScalarField object (Ms) found with name "
      + (*FindInitValue("Ms"))[0];
    throw Oxs_Ext::Error(msg.c_str());
  }
  DeleteInitValue("Ms");

  CheckInitValueParamCount("m0",1);
  Oxs_DepKey<Oxs_Ext> key3
    = director->FindExtObject((*FindInitValue("m0"))[0]);
  Oxs_FixedVectorField* m0
    = dynamic_cast<Oxs_FixedVectorField*>(key3.GetPtr());
  if(m0==NULL) {
    string msg="No Oxs_FixedVectorField object (m0) found with name "
      + (*FindInitValue("m0"))[0];
    throw Oxs_Ext::Error(msg.c_str());
  }
  DeleteInitValue("m0");

  VerifyAllInitArgsUsed();

  // Setup initial state.
  // The following elements should eventually be wrapped up in a
  // STL map array, and indexed by name.
  director->ReserveSimulationStateRequest(1);
  director->GetNewSimulationState(initial_state);
  Oxs_SimState& istate = initial_state.GetWriteReference();
  istate.last_timestep   =  0.0;
  istate.elapsed_time    =  0.0;
  istate.max_dm_dt       = -1.0;     // Dummy value.
  istate.total_energy    =  0.0;     // Dummy value.
  istate.iteration_count =  0;
  istate.stage_iteration_count =  0;
  istate.stage_number    =  0;

  istate.Ms = Ms;
  istate.mesh = mesh;
  istate.spin.AdjustSize(mesh);

  UINT4m i,size=mesh->Size();
  for(i=0;i<size;i++) istate.spin[i]=(*m0)[i];

  istate.valid_state=0;  // Technically, this is not a valid
  /// state because max_dm_dt and total_energy aren't set.

  initial_state.GetReadReference();
}

BOOL Oxs_BaseDriver::Init()
{ 
  Init_Driver();  // Run init routine in parent
  return 1;
}

Oxs_BaseDriver::~Oxs_BaseDriver()
{}

BOOL Oxs_BaseDriver::StageDone(const Oxs_SimState& state,
			       Tcl_Interp *interp,
			       int /* argc */, char** /* argv */) const
{ // Note: The iteration counts are effective 1-based, because
  // the original state in each stage is marked 'invalid_state,'
  // since max_dm_dt & energy aren't set.  Thus, since state
  // iteration counts record the count for a completed state,
  // when the iteration count is the same as the limit, the limit
  // is reached.
  if( stage_iteration_limit>0
      && state.stage_iteration_count >= stage_iteration_limit) goto done;
  if( total_iteration_limit>0
      && state.iteration_count >= total_iteration_limit ) goto done;
  if( !state.valid_state ) return 0;
  if( state.max_dm_dt <= stopping_dm_dt ) goto done;
  return 0;
done:
  Oc_AutoBuf ab;
  Oc_Snprintf(buf, sizeof(buf), 
	      "Oc_EventHandler Generate Oxs_Run StageDone -stage %d",
	      state.stage_number);
  Tcl_Eval(interp, ab(buf));
  return 1;
}

BOOL Oxs_BaseDriver::Done(const Oxs_SimState& state,
			  Tcl_Interp *interp,
			  int /* argc */, char** /* argv */) const
{ // Note: See note on iteration counts in StageDone() above.
  if( total_iteration_limit>0
      && state.iteration_count >= total_iteration_limit ) goto done;
  if( number_of_stages >0
      && state.stage_number >= number_of_stages ) goto done;
  return 0;
done:
  Oc_AutoBuf ab;
  Tcl_Eval(interp, ab("Oc_EventHandler Generate Oxs_Run Done"));
  return 1;
}

BOOL Oxs_BaseDriver::Return(const Oxs_SimState& /* state */,
			    Tcl_Interp* /* interp */,
			    int /* argc */, char** /* argv */) const
{
  return 1;
}

void Oxs_BaseDriver::NewState(const Oxs_SimState& state,
			      Tcl_Interp *interp,
			      int /* argc */, char** /* argv */) const
{
  Oc_AutoBuf ab;
  Oc_Snprintf(buf, sizeof(buf), 
              "Oc_EventHandler Generate Oxs_Run NewState -step %d",
              state.iteration_count);
  Tcl_Eval(interp, ab(buf));
}

void Oxs_BaseDriver::FillStatePartA
(const Oxs_SimState& old_state,
 Oxs_SimState& new_state) const
{
  new_state.valid_state  = 0;
  new_state.stage_number = old_state.stage_number;
  new_state.Ms           = old_state.Ms;
  new_state.mesh         = old_state.mesh;
  new_state.spin.AdjustSize(new_state.mesh);
}

void Oxs_BaseDriver::FillStatePartB(Oxs_SimState& work_state) const
{
  if(work_state.last_timestep<min_timestep)
    work_state.last_timestep=min_timestep;
  if(work_state.last_timestep>max_timestep)
    work_state.last_timestep=max_timestep;
}
