/* FILE: cubicdemag.h            -*-Mode: c++-*-
 *
 * Average H demag field across cubic cells.
 *
 */

#include "nb.h"
#include "depkey.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

#include "cubicmesh.h"
#include "cubicdemag.h"
#include "demagcoef.h"
#include "fft.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_CubicDemag);

/* End includes */

// Helper function
UINT4m Oxs_CubicDemag::NextPowerOfTwo(UINT4m n) const
{ // Returns first power of two >= n
  UINT4m m=1;
  while(m<n) m*=2;
  return m;
}

// Constructor
Oxs_CubicDemag::Oxs_CubicDemag(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr),
    xdim(0),ydim(0),zdim(0),totalsize(0),
    pxdim(0),pydim(0),pzdim(0),ptotalsize(0),
    A00(NULL),A01(NULL),A02(NULL),A11(NULL),A12(NULL),A22(NULL),
    Mx(NULL),My(NULL),Mz(NULL),Hcomp(NULL)
{

#ifdef DEBUGTEST
  // 3D FFT Debug test
#define XS 2
#define YS 4
#define ZS 8
  UINT4m totalsize=XS*YS*ZS;
  MyComplex* arr = new MyComplex[totalsize];
  for(UINT4m i=0;i<totalsize;i++) {
    arr[i]=MyComplex(i,i+1);
  }

  printf("ORIGINAL DATA -----------------------\n");
  for(int k=ZS-1;k>=0;k--) {
    for(int j=YS-1;j>=0;j--) {
      for(int i=0;i<XS;i++) {
	printf(" a[%d][%2d][%d]=(%5.2f,%5.2f)",
	       i,j,k,arr[i+j*XS+k*XS*YS].real(),arr[i+j*XS+k*XS*YS].imag());
      }
      printf("\n");
    }
    printf("\n");
  }

  FFT3D fft;
  fft.Forward(XS,YS,ZS,arr);
  printf("TRANSFORMED DATA -----------------------\n");
  for(int k=ZS-1;k>=0;k--) {
    for(int j=YS-1;j>=0;j--) {
      for(int i=0;i<XS;i++) {
	printf(" a[%d][%2d][%d]=(%5.2f,%5.2f)",
	       i,j,k,arr[i+j*XS+k*XS*YS].real(),arr[i+j*XS+k*XS*YS].imag());
      }
      printf("\n");
    }
    printf("\n");
  }

  fft.Inverse(XS,YS,ZS,arr);
  printf("FORWARD+INVERSE DATA -----------------------\n");
  for(int k=ZS-1;k>=0;k--) {
    for(int j=YS-1;j>=0;j--) {
      for(int i=0;i<XS;i++) {
	printf(" a[%d][%2d][%d]=(%5.2f,%5.2f)",
	       i,j,k,arr[i+j*XS+k*XS*YS].real(),arr[i+j*XS+k*XS*YS].imag());
      }
      printf("\n");
    }
    printf("\n");
  }
  delete[] arr;
#endif // DEBUGTEST

  CheckInitValueParamCount("mesh",1);
  const string& meshname = (*FindInitValue("mesh"))[0];
  Oxs_DepKey<Oxs_Ext> tempkey = director->FindExtObject(meshname);
  if(tempkey.GetPtr()==NULL) {
    string msg="No Oxs_Mesh object found with name " + meshname;
    throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_CubicMesh* mesh = dynamic_cast<Oxs_CubicMesh*>(tempkey.GetPtr());
  if(mesh==NULL) {
    string msg="Object " + meshname + " is not a cubic mesh.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  meshkey.Set(mesh); // Dependency lock is held until *this is deleted.
  DeleteInitValue("mesh");

  VerifyAllInitArgsUsed();

  // Fill dimension variables
  xdim = mesh->XDim();
  ydim = mesh->YDim();
  zdim = mesh->ZDim();
  totalsize=xdim*ydim*zdim;
  if(xdim==0 || ydim==0 || zdim==0) return; // Empty mesh!
  if(totalsize < xdim || totalsize < ydim || totalsize < zdim) {
    // Partial overflow check
    string msg="UINT4m overflow in " + string(InstanceName())
      + ": Product xdim*ydim*zdim too big to fit in a UINT4m variable";
    throw Oxs_Ext::Error(msg.c_str());
  }
  pxdim = NextPowerOfTwo(2*xdim);
  pydim = NextPowerOfTwo(2*ydim);
  pzdim = NextPowerOfTwo(2*zdim);
  ptotalsize=pxdim*pydim*pzdim;
  if(ptotalsize < pxdim || ptotalsize < pydim || ptotalsize < pzdim) {
    // Partial overflow check
    string msg="UINT4m overflow in " + string(InstanceName())
      + ": Product pxdim*pydim*pzdim too big to fit in a UINT4m variable";
    throw Oxs_Ext::Error(msg.c_str());
  }


  // Allocate memory for interaction matrices and magnetization components
  A00   = new MyComplex[ptotalsize];
  A01   = new MyComplex[ptotalsize];
  A02   = new MyComplex[ptotalsize];
  A11   = new MyComplex[ptotalsize];
  A12   = new MyComplex[ptotalsize];
  A22   = new MyComplex[ptotalsize];
  Mx    = new MyComplex[ptotalsize];
  My    = new MyComplex[ptotalsize];
  Mz    = new MyComplex[ptotalsize];
  Hcomp = new MyComplex[ptotalsize];
  if(Hcomp==NULL || Mx==NULL || My==NULL || Mz==NULL
     || A00==NULL || A01==NULL || A02==NULL
     || A11==NULL || A12==NULL || A22==NULL) {
    // Safety check for those machines on which new[] doesn't throw
    // BadAlloc.
    string msg="Insufficient memory in cubicdemag constructor.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Initialize interaction matrices to zeros
  UINT4m pindex;
  for(pindex=0;pindex<ptotalsize;pindex++) A00[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A01[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A02[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A11[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A12[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) A22[pindex].Set(0.,0.);

  // According (16) in Newell's paper, the demag field is given by
  //                        H = -N*M
  // where N is the "demagnetizing tensor," with components Nxx, Nxy,
  // etc.  With the '-1' in 'scale' we store '-N' instead of 'N',
  // so we don't have to multiply the output from the FFT + iFFT
  // by -1 in ConstMagField() below.
  REALWIDE scale = -1./(4*PI);  // Assumes cubes

  // Fill interaction matrices with demag coefs from Newell's paper.
  // Note that A00, A11 and A22 are even in x,y and z.
  // A01 is odd in x and y, even in z.
  // A02 is odd in x and z, even in y.
  // A12 is odd in y and z, even in x.
  UINT4m i,j,k;
  UINT4m pxydim=pxdim*pydim;
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE coef;
    coef=scale*CalculateSDA00(i,j,k,1,1,1);
    A00[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A00[(pxdim-i)+j*pxdim+k*pxydim].Set(coef,0.);
    if(j>0) A00[i+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(k>0) A00[i+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0)
      A00[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(i>0 && k>0)
      A00[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(j>0 && k>0)
      A00[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0 && k>0)
      A00[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE coef;
    coef=scale*CalculateSDA11(i,j,k,1,1,1);
    A11[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A11[(pxdim-i)+j*pxdim+k*pxydim].Set(coef,0.);
    if(j>0) A11[i+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(k>0) A11[i+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0)
      A11[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(i>0 && k>0)
      A11[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(j>0 && k>0)
      A11[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0 && k>0)
      A11[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE coef;
    coef=scale*CalculateSDA22(i,j,k,1,1,1);
    A22[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A22[(pxdim-i)+j*pxdim+k*pxydim].Set(coef,0.);
    if(j>0) A22[i+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(k>0) A22[i+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0)
      A22[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(i>0 && k>0)
      A22[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(j>0 && k>0)
      A22[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0 && k>0)
      A22[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE coef;
    coef=scale*CalculateSDA01(i,j,k,1,1,1);
    A01[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A01[(pxdim-i)+j*pxdim+k*pxydim].Set(-coef,0.);
    if(j>0) A01[i+(pydim-j)*pxdim+k*pxydim].Set(-coef,0.);
    if(k>0) A01[i+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0)
      A01[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(i>0 && k>0)
      A01[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(j>0 && k>0)
      A01[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(i>0 && j>0 && k>0)
      A01[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE coef;
    coef=scale*CalculateSDA02(i,j,k,1,1,1);
    A02[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A02[(pxdim-i)+j*pxdim+k*pxydim].Set(-coef,0.);
    if(j>0) A02[i+(pydim-j)*pxdim+k*pxydim].Set(coef,0.);
    if(k>0) A02[i+j*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(i>0 && j>0)
      A02[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(-coef,0.);
    if(i>0 && k>0)
      A02[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(j>0 && k>0)
      A02[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(i>0 && j>0 && k>0)
      A02[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE coef;
    coef=scale*CalculateSDA12(i,j,k,1,1,1);
    A12[i+j*pxdim+k*pxydim].Set(coef,0.);
    if(i>0) A12[(pxdim-i)+j*pxdim+k*pxydim].Set(coef,0.);
    if(j>0) A12[i+(pydim-j)*pxdim+k*pxydim].Set(-coef,0.);
    if(k>0) A12[i+j*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(i>0 && j>0)
      A12[(pxdim-i)+(pydim-j)*pxdim+k*pxydim].Set(-coef,0.);
    if(i>0 && k>0)
      A12[(pxdim-i)+j*pxdim+(pzdim-k)*pxydim].Set(-coef,0.);
    if(j>0 && k>0)
      A12[i+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
    if(i>0 && j>0 && k>0)
      A12[(pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim].Set(coef,0.);
  }

  // Transform interaction matrices into frequency domain.
  fft.Forward(pxdim,pydim,pzdim,A00);
  fft.Forward(pxdim,pydim,pzdim,A01);
  fft.Forward(pxdim,pydim,pzdim,A02);
  fft.Forward(pxdim,pydim,pzdim,A11);
  fft.Forward(pxdim,pydim,pzdim,A12);
  fft.Forward(pxdim,pydim,pzdim,A22);
}

void Oxs_CubicDemag::ReleaseMemory()
{
  if(Mx!=NULL) { delete[] Mx; Mx=NULL; }
  if(My!=NULL) { delete[] My; My=NULL; }
  if(Mz!=NULL) { delete[] Mz; Mz=NULL; }
  if(A00!=NULL) { delete[] A00; A00=NULL; }
  if(A01!=NULL) { delete[] A01; A01=NULL; }
  if(A02!=NULL) { delete[] A02; A02=NULL; }
  if(A11!=NULL) { delete[] A11; A11=NULL; }
  if(A12!=NULL) { delete[] A12; A12=NULL; }
  if(A22!=NULL) { delete[] A22; A22=NULL; }
  xdim=ydim=zdim=totalsize=0;
  pxdim=pydim=pzdim=ptotalsize=0;
}

BOOL Oxs_CubicDemag::Init()
{
  return 1;
}

void Oxs_CubicDemag::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  // Check that the mesh hasn't changed.
  const Oxs_CubicMesh* mesh = meshkey.GetPtr();
  if(mesh != state.mesh) {
    string msg="Input mesh to Oxs_CubicDemag::GetEnergyAndField()"
      " routine of object " +  string(InstanceName())
      + " is different than the one used to initialize this object.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_FixedScalarField& Ms = *(state.Ms);

  // Calculate FFT of Mx, My and Mz
  UINT4m xydim=xdim*ydim;
  UINT4m pxydim=pxdim*pydim;
  UINT4m index,pindex;
  UINT4m i,j,k;
  for(pindex=0;pindex<ptotalsize;pindex++) Mx[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) My[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) Mz[pindex].Set(0.,0.);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    REAL8m scale = Ms[index];
    const ThreeVector& vec = spin[index];
    Mx[pindex].Set(scale*(vec.x),0);
    My[pindex].Set(scale*(vec.y),0);
    Mz[pindex].Set(scale*(vec.z),0);
  }
  fft.Forward(pxdim,pydim,pzdim,Mx);
  fft.Forward(pxdim,pydim,pzdim,My);
  fft.Forward(pxdim,pydim,pzdim,Mz);
  
  // Calculate field components in frequency domain, then to iFFT
  // to transform back to space domain.

  // Hx component
  for(pindex=0;pindex<ptotalsize;pindex++) {
    Hcomp[pindex] = A00[pindex]*Mx[pindex]
      + A01[pindex]*My[pindex] + A02[pindex]*Mz[pindex];
  }
  fft.Inverse(pxdim,pydim,pzdim,Hcomp);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    field[index].x=Hcomp[pindex].real();
  }
  
  // Hy component
  for(pindex=0;pindex<ptotalsize;pindex++) {
    Hcomp[pindex] = A01[pindex]*Mx[pindex]
      + A11[pindex]*My[pindex] + A12[pindex]*Mz[pindex];
  }
  fft.Inverse(pxdim,pydim,pzdim,Hcomp);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    field[index].y=Hcomp[pindex].real();
  }
  
  // Hz component
  for(pindex=0;pindex<ptotalsize;pindex++) {
    Hcomp[pindex] = A02[pindex]*Mx[pindex]
      + A12[pindex]*My[pindex] + A22[pindex]*Mz[pindex];
  }
  fft.Inverse(pxdim,pydim,pzdim,Hcomp);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    field[index].z=Hcomp[pindex].real();
  }

  // Calculate pointwise energy density: -0.5*MU0*<M,H>
  REAL8m mult = -0.5 * MU0;
  for(index=0;index<totalsize;index++) {
    energy[index] = mult * Ms[index] * (spin[index]*field[index]);
  }

}
