/* FILE: cubicmesh.cc         -*-Mode: c++-*-
 *
 * Cubic mesh, derived from Oxs_Mesh class.
 *
 */

#include "vf.h"
#include "meshvalue.h"
#include "cubicmesh.h"
#include "depkey.h"
#include "director.h"
#include "nb.h"
#include "rectangularregion.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy
#include "util.h"


// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_CubicMesh);

/* End includes */

/////////////////////////////////////////////////////////////////////
// Oxs_CubicMesh
Oxs_CubicMesh::Oxs_CubicMesh
(const char* name,     // Child instance id
 Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
 const char* argstr)   // MIF input block parameters
  : Oxs_Mesh(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  CheckInitValueParamCount("cellsize",1);
  cellsize=Nb_Atof((*FindInitValue("cellsize"))[0].c_str());
  DeleteInitValue("cellsize");

  CheckInitValueParamCount("region",1);
  const Oxs_RectangularRegion* region=NULL;
  Oxs_DepKey<Oxs_Ext> key
    = director->FindExtObject((*FindInitValue("region"))[0]);
  if(key.GetPtr()!=NULL) {
    region = dynamic_cast<const Oxs_RectangularRegion*>(key.GetPtr());
    // Note: The dependency lock will be released when key
    // is destroyed, i.e., when this function returns.
  }
  DeleteInitValue("region");

  if(cellsize<=0.0 || region==NULL) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  VerifyAllInitArgsUsed();

  Oxs_Box box;
  region->GetExtents(box);
  REAL8m xrange = box.GetMaxX() - box.GetMinX();
  REAL8m yrange = box.GetMaxY() - box.GetMinY();
  REAL8m zrange = box.GetMaxZ() - box.GetMinZ();

  base = ThreeVector(box.GetMinX()+cellsize/2.,
		     box.GetMinY()+cellsize/2.,
		     box.GetMinZ()+cellsize/2.);

  xdim = static_cast<int>(OC_ROUND(xrange/cellsize));
  ydim = static_cast<int>(OC_ROUND(yrange/cellsize));
  zdim = static_cast<int>(OC_ROUND(zrange/cellsize));
  if(xdim<1 || ydim<1 || zdim<1) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  xydim = xdim*ydim;
  elementcount = xydim*zdim;
  if(elementcount<xdim || elementcount<ydim || elementcount<zdim) {
    /// The above test is not complete, but will catch
    /// some overflows.
    string msg="UINT4m overflow: mesh size too large in "
      "Oxs_CubicMesh constructor";
    throw Oxs_Ext::Error(msg.c_str());
  }

  if( fabs(xdim*cellsize - xrange) > 0.01*cellsize ||
      fabs(ydim*cellsize - yrange) > 0.01*cellsize ||
      fabs(zdim*cellsize - zrange) > 0.01*cellsize   ) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName())
      + ": range is not an integral multiple of cellsize.";
    throw Oxs_Ext::Error(msg.c_str());
  }
}


void Oxs_CubicMesh::GetBoundingBox(Oxs_Box& bbox) const
{
  bbox.Set(base.x-cellsize/2., base.x-cellsize/2.+xdim*cellsize,
	   base.y-cellsize/2., base.y-cellsize/2.+ydim*cellsize,
	   base.z-cellsize/2., base.z-cellsize/2.+zdim*cellsize);
}

void Oxs_CubicMesh::Center(UINT4m index,ThreeVector &location) const
{
#ifndef NDEBUG
  if(index>elementcount) {
    string msg="Index out of range "
      "(Oxs_CubicMesh::Location(INT4m) const)";
    throw Oxs_Ext::Error(msg.c_str());
  }
#endif
  INT4m iz = index/(xdim*ydim);  index -= iz*xdim*ydim;
  INT4m iy = index/xdim;         index -= iy*xdim;
  INT4m ix = index;
  location.Set(base.x+ix*cellsize,
               base.y+iy*cellsize,
               base.z+iz*cellsize);
  return;
}




// File output.  Returns 0 on success.
int Oxs_CubicMesh::WriteOvf
(const char* filename,  // Name of output file
 const char* title,     // Long filename or title
 const char* desc,      // Description to embed in output file
 const char* meshtype,  // Either "rectangular" or "irregular"
 const char* datatype,  // Either "binary" or "text"
 const char* precision, // For binary, "4" or "8";
                        //  for text, a printf-style format
 const Oxs_MeshValue<ThreeVector>* vec,  // Vector array
 const Oxs_MeshValue<REAL8m>* scale      // Optional scaling for vec
 /// Set scale to NULL to use vec values directly.
 ) const
{
  // Use default file writer for irregular mesh.
  if(strcmp("irregular",meshtype)==0) {
    ThreeVector stephints(cellsize,cellsize,cellsize);
    return WriteOvfIrregular(filename,title,desc,datatype,precision,
			     vec,scale,&stephints);
  }

  if(strcmp("rectangular",meshtype)!=0) {
    string msg=string("Unrecognized mesh type request: ")
      + string(meshtype)
      + string("(Oxs_CubicMesh::WriteOvf() in object")
      + string(InstanceName()) + string(")");
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Rectangular Mesh ////////////////////////////
  // Check import validity
  enum DataType { BINARY, TEXT };
  DataType dt = BINARY;
  const char* filemode = "wb";
  if(strcmp("text",datatype)==0) {
    dt = TEXT;
    filemode = "w";
  } else if(strcmp("binary",datatype)!=0){
    return 1;  // Bad datatype
  }

  int datawidth = 0;
  string dataformat;
  if(dt == BINARY) {
    datawidth = atoi(precision);
    if(datawidth != 4 && datawidth != 8) {
      return 2; // Bad precision
    }
  } else {
    if(precision==NULL || precision[0]=='\0')
      precision="%# .17g";  // Default format
    string temp = string(precision) + string(" ");
    dataformat = temp + temp + string(precision) + string("\n");
  }

  if(!vec->CheckMesh(this) || (scale!=NULL && scale->CheckMesh(this))) {
    return 3; // Import data not compatible with *this
  }

  // Open file
  Oxs_FilePointer file(filename,filemode);

  // Write header
  fprintf(file.fptr,"# OOMMF: rectangular mesh v1.0\n");
  fprintf(file.fptr,"# Segment count: 1\n");
  fprintf(file.fptr,"# Begin: Segment\n");
  fprintf(file.fptr,"# Begin: Header\n");
  if(title==NULL || title[0]=='\0')
    fprintf(file.fptr,"# Title: %s\n",filename);
  else
    fprintf(file.fptr,"# Title: %s\n",title);
  if(desc!=NULL && desc[0]!='\0') {
    // Print out description, breaking at newlines as necessary.
    // Note: This block is optional
    const char *cptr1,*cptr2;
    cptr1=desc;
    while((cptr2=strchr(cptr1,'\n'))!=NULL) {
      fprintf(file.fptr,"# Desc: %.*s\n",(int)(cptr2-cptr1),cptr1);
      cptr1=cptr2+1;
    }
    if(*cptr1!='\0') fprintf(file.fptr,"# Desc: %s\n",cptr1);
  }
  fprintf(file.fptr,"# meshunit: m\n");
  fprintf(file.fptr,"# valueunit: A/m\n");
  fprintf(file.fptr,"# valuemultiplier: 1.0\n");

  Oxs_Box bbox;
  GetBoundingBox(bbox);
  fprintf(file.fptr,
	  "# xmin: %.16g\n# ymin: %.16g\n# zmin: %.16g\n"
	  "# xmax: %.16g\n# ymax: %.16g\n# zmax: %.16g\n",
	  bbox.GetMinX(),bbox.GetMinY(),bbox.GetMinZ(),
	  bbox.GetMaxX(),bbox.GetMaxY(),bbox.GetMaxZ());

  // Skip "boundary:" for now

  REAL8m minmag=0,maxmag=0;
  if(Size()>0) {
    REAL8m tempscale=1.0;
    maxmag=(*vec)[0].MagSq();
    if(scale!=NULL) {
      tempscale=(*scale)[0];
      maxmag*=tempscale*tempscale;
    }
    minmag=maxmag;
    for(UINT4m i=1;i<Size();i++) {
      REAL8m val=(*vec)[i].MagSq();
      if(scale!=NULL) {
	tempscale=(*scale)[i];
	val*=tempscale*tempscale;
      }
      if(val<minmag) minmag=val;
      if(val>maxmag) maxmag=val;
    }
    maxmag=sqrt(maxmag);
    minmag=sqrt(minmag);
    minmag*=0.999;  // Underestimate lower bound by 0.1% to protect against
    /// rounding errors.
  }

  fprintf(file.fptr,"# ValueRangeMaxMag: %.16g\n",maxmag);
  fprintf(file.fptr,"# ValueRangeMinMag: %.16g\n",minmag*1e-8);
  /// minmag is multiplied by 1e-8 to specify contrast range.

  fprintf(file.fptr,"# meshtype: rectangular\n");
  fprintf(file.fptr,"# xbase: %.16g\n# ybase: %.16g\n# zbase: %.16g\n",
	  base.x,base.y,base.z);
  fprintf(file.fptr,
	  "# xstepsize: %.16g\n# ystepsize: %.16g\n# zstepsize: %.16g\n",
	  cellsize,cellsize,cellsize);
  fprintf(file.fptr,"# xnodes: %d\n# ynodes: %d\n# znodes: %d\n",
	  xdim,ydim,zdim);

  fprintf(file.fptr,"# End: Header\n");

  // Write data block
  if(dt == BINARY) {
    if(datawidth==4) {
      fprintf(file.fptr,"# Begin: Data Binary 4\n");
      REAL4 buf[3];
      buf[0]=1234567.;  // 4-Byte checkvalue
      Vf_OvfFileFormatSpecs::WriteBinary(file.fptr,buf,1);
      INT4m size=Size();
      for(INT4m i=0 ; i<size ; ++i) {
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  buf[0] = REAL4(v.x);
	  buf[1] = REAL4(v.y);
	  buf[2] = REAL4(v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  buf[0] = REAL4(tempscale*v.x);
	  buf[1] = REAL4(tempscale*v.y);
	  buf[2] = REAL4(tempscale*v.z);
	}
        Vf_OvfFileFormatSpecs::WriteBinary(file.fptr,buf,3);
	// Vf_OvfFileFormatSpecs::WriteBinary performs
	// byte-swapping as needed.
      }
      fprintf(file.fptr,"\n# End: Data Binary 4\n");
    } else {
      fprintf(file.fptr,"# Begin: Data Binary 8\n");
      REAL8 buf[3];
      buf[0]=123456789012345.;  // 8-Byte checkvalue
      Vf_OvfFileFormatSpecs::WriteBinary(file.fptr,buf,1);
      INT4m size=Size();
      for(INT4m i=0 ; i<size ; ++i) {
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  buf[0] = REAL4(v.x);
	  buf[1] = REAL4(v.y);
	  buf[2] = REAL4(v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  buf[0] = REAL4(tempscale*v.x);
	  buf[1] = REAL4(tempscale*v.y);
	  buf[2] = REAL4(tempscale*v.z);
	}
        Vf_OvfFileFormatSpecs::WriteBinary(file.fptr,buf,3);
	// Vf_OvfFileFormatSpecs::WriteBinary performs
	// byte-swapping as needed.
      }
      fprintf(file.fptr,"\n# End: Data Binary 8\n");
    }
  } else {
    fprintf(file.fptr,"# Begin: Data Text\n");
    INT4m size=Size();
    for(INT4m i=0 ; i<size ; ++i) {
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  fprintf(file.fptr,dataformat.c_str(),
		  v.x,v.y,v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  fprintf(file.fptr,dataformat.c_str(),
		  tempscale*v.x,tempscale*v.y,tempscale*v.z);
	}
    }
    fprintf(file.fptr,"# End: Data Text\n");
  }
  fprintf(file.fptr,"# End: Segment\n");
  return 0;
}
