/* FILE: fft.h                    -*-Mode: c++-*-
 *
 *	C++ header file for 1 and 2 dimensional FFT code.
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2000/08/17 00:15:38 $
 * Last modified by: $Author: donahue $
 */

#ifndef _FFT_H
#define _FFT_H

/* End includes */

typedef REALWIDE FFT_REAL_TYPE;

#define USE_MYCOMPLEX

#ifdef USE_MYCOMPLEX
typedef FFT_REAL_TYPE MY_COMPLEX_REAL_TYPE;
class MyComplex {
public:
  MY_COMPLEX_REAL_TYPE re,im;
  MyComplex() { re=im=0.0; }
  MyComplex(MY_COMPLEX_REAL_TYPE x,MY_COMPLEX_REAL_TYPE y) { re=x; im=y; }
  void Set(MY_COMPLEX_REAL_TYPE x,MY_COMPLEX_REAL_TYPE y) { re=x; im=y; }
  friend inline MyComplex operator+(const MyComplex& a,const MyComplex& b);
  friend inline MyComplex operator-(const MyComplex& a,const MyComplex& b);
  friend inline MyComplex operator*(const MyComplex& a,const MyComplex& b);
  friend inline MyComplex operator*(const MyComplex& a,
				    MY_COMPLEX_REAL_TYPE b);
  friend inline MyComplex operator*(MY_COMPLEX_REAL_TYPE b,MyComplex& a);
  MyComplex& operator+=(const MyComplex& b);
  MyComplex& operator*=(const MyComplex& b);
  MyComplex& operator*=(MY_COMPLEX_REAL_TYPE b);
  inline MY_COMPLEX_REAL_TYPE real() const { return re; }
  inline MY_COMPLEX_REAL_TYPE imag() const { return im; }
};

inline MyComplex operator*(const MyComplex& a,const MyComplex& b)
{ return MyComplex(a.re*b.re-a.im*b.im,a.re*b.im+a.im*b.re); }

inline MyComplex operator+(const MyComplex& a,const MyComplex& b)
{ return MyComplex(a.re+b.re,a.im+b.im); }

inline MyComplex operator-(const MyComplex& a,const MyComplex& b)
{ return MyComplex(a.re-b.re,a.im-b.im); }

inline MyComplex operator*(const MyComplex& a,MY_COMPLEX_REAL_TYPE b)
{ return MyComplex(a.re*b,a.im*b); }

inline MyComplex operator*(MY_COMPLEX_REAL_TYPE b,MyComplex& a)
{ return MyComplex(a.re*b,a.im*b); }

inline MyComplex& MyComplex::operator+=(const MyComplex& b)
{ re+=b.re; im+=b.im; return *this; }

inline MyComplex& MyComplex::operator*=(const MyComplex& b)
{
  MY_COMPLEX_REAL_TYPE temp;
  re=(temp=re)*b.re-im*b.im; im=im*b.re+temp*b.im;
  return *this;
}

inline MyComplex& MyComplex::operator*=(MY_COMPLEX_REAL_TYPE b)
{ re*=b; im*=b; return *this; }

/* typedef MyComplex Complex; */
/* NOTE: <X11/X.h> has a "#define Complex 0" line
 *    (apparently associated with defining polygon
 *    types) that can interfere with the preceding
 *    typedef.  So presently (28-April-97) I've (mjd)
 *    decided to use MyComplex throughout.  If
 *    nothing else, this should make it easier to
 *    modify Complex type usage later on.  See, for
 *    example, the typedef in the !USE_MYCOMPLEX
 *    #ifdef branch below.
 */
#else  /* !USE_MYCOMPLEX */
typedef Complex MyComplex;
#endif /* USE_MYCOMPLEX */

inline MyComplex conj(const MyComplex& a)
{
  return MyComplex(a.real(),-a.imag());
}

class FFT {
  // NOTE: All member functions except for constructor and
  // destructor are "conceptually const"
private:
  mutable int vecsize;
  enum Direction { forward, inverse };
  mutable MyComplex *Uforward,*Uinverse;
  mutable MyComplex *workspace;
  mutable int *permindex;
  void Permute(MyComplex *vec) const;
  void BaseDecFreqForward(MyComplex *vec) const;
  void BaseDecTimeInverse(MyComplex *vec) const;
  void Setup(int size) const;
public:
  FFT() {
    vecsize=0;
    Uforward=Uinverse=workspace=(MyComplex *)NULL;
    permindex=(int *)NULL;
  }
  ~FFT();
  void ReleaseMemory() const; // Conceptually const!?
  void ForwardDecFreq(int size,MyComplex *vec,
		      FFT_REAL_TYPE divisor=0.) const;
  void InverseDecTime(int size,MyComplex *vec,
		      FFT_REAL_TYPE divisor=0.) const;
};


// Class for in-place 3D complex FFT's
class FFT3D {
  // NOTE: All member functions except for constructor and
  // destructor are "conceptually const"
  mutable UINT4m scratchsize;
  mutable MyComplex* scratch;  // Buffer space large enough for any single dim
  FFT xfft,yfft,zfft;
  void ReleaseMemory() const;
  void AdjustScratchSize(UINT4m xsize,UINT4m ysize,UINT4m zsize) const;
public:
  FFT3D() : scratchsize(0), scratch(NULL) {}
  ~FFT3D() { ReleaseMemory(); }
  // This code assumes arr is in fortran order, i.e.,
  // increment first x, then y, and last z.
  void Forward(UINT4m xsize,UINT4m ysize,UINT4m zsize,
	       MyComplex* arr) const;
  void Inverse(UINT4m xsize,UINT4m ysize,UINT4m zsize,
	       MyComplex* arr) const;
};

#endif /* _FFT_H */
