/* FILE: myanisotropy.cc            -*-Mode: c++-*-
 *
 * My Uniform Uniaxial Anisotropy, derived from Oxs_Energy class.
 *
 */

#include "oc.h"
#include "nb.h"
#include "threevector.h"
#include "director.h"
#include "simstate.h"
#include "ext.h"
#include "depkey.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "cubicmesh.h"
#include "myanisotropy.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(My_Anisotropy);

/* End includes */


// Constructor
My_Anisotropy::My_Anisotropy(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  CheckInitValueParamCount("K1",1);
  K1=Nb_Atof((*FindInitValue("K1"))[0].c_str());
  DeleteInitValue("K1");

  CheckInitValueParamCount("Ms",1);
  Ms=Nb_Atof((*FindInitValue("Ms"))[0].c_str());
  DeleteInitValue("Ms");

  CheckInitValueParamCount("axis1",3);
  axis1.x=Nb_Atof((*FindInitValue("axis1"))[0].c_str());
  axis1.y=Nb_Atof((*FindInitValue("axis1"))[1].c_str());
  axis1.z=Nb_Atof((*FindInitValue("axis1"))[2].c_str());
  DeleteInitValue("axis1");

  CheckInitValueParamCount("axis2",3);
  axis2.x=Nb_Atof((*FindInitValue("axis2"))[0].c_str());
  axis2.y=Nb_Atof((*FindInitValue("axis2"))[1].c_str());
  axis2.z=Nb_Atof((*FindInitValue("axis2"))[2].c_str());
  DeleteInitValue("axis2");

  VerifyAllInitArgsUsed();

  if(Ms==0.0) {
    K1=0.0; Ms=1.0; // Safety
  }

  REAL8m magsq=axis1.MagSq();
  if(magsq==0.0) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName())
      + ": Specified anisotropy axis 1 is (0,0,0)";
    throw Oxs_Ext::Error(msg.c_str());
  }
  if(magsq!=1.0) axis1 *= 1.0/sqrt(magsq);

  magsq=axis2.MagSq();
  if(magsq==0.0) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName())
      + ": Specified anisotropy axis 2 is (0,0,0)";
    throw Oxs_Ext::Error(msg.c_str());
  }
  if(magsq!=1.0) axis2 *= 1.0/sqrt(magsq);

  if(fabs(axis1*axis2)>1e-12) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName())
      + ": Specified anisotropy axes aren't perpendicular";
    throw Oxs_Ext::Error(msg.c_str());
  }

}

BOOL My_Anisotropy::Init()
{ return 1; }

void My_Anisotropy::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  UINT4m size = state.mesh->Size();
  ThreeVector axis3 = axis1 ^ axis2;
  axis3.SetMag(1.0); // Just to be safe

  REAL8m field_mult = -2*K1/(Ms*MU0);
  for(UINT4m i=0;i<size;++i) {
    REAL8m a = axis1*state.spin[i];
    REAL8m b = axis2*state.spin[i];
    REAL8m c = axis3*state.spin[i];
    energy[i] = K1 * (a*a*b*b+a*a*c*c+b*b*c*c);
    field[i]  = (a*(b*b+c*c))*axis1;
    field[i] += (b*(a*a+c*c))*axis2;
    field[i] += (c*(a*a+b*b))*axis3;
    field[i] *= field_mult;
  }

}
