/* FILE: tsfvectorfield.cc      -*-Mode: c++-*-
 *
 * Tcl Script Fixed Vector Field, derived from Oxs_FixedVectorField class.
 *
 */

#include <string>

#include "oc.h"
#include "nb.h"

#include "util.h"
#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "tsfvectorfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_TSFVectorField);

/* End includes */


// Constructor
Oxs_TSFVectorField::Oxs_TSFVectorField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_FixedVectorField(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  BOOL setnorm = 0;
  REAL8m norm=0;
  const vector<string>* normlist = FindInitValue("norm");
  if(normlist!=NULL) {
    setnorm = 1;
    CheckInitValueParamCount("norm",1);
    norm = Nb_Atof((*normlist)[0].c_str());
    DeleteInitValue("norm");
  }

  const vector<string>* scriptlist = FindInitValue("script");
  if(scriptlist==NULL) {
    string msg="No script specified in input block for object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  string script=(*scriptlist)[0];
  for(UINT4m iarg=1;iarg<scriptlist->size();iarg++) {
    script += " ";
    script += (*scriptlist)[iarg];
  }
  DeleteInitValue("script");

  CheckInitValueParamCount("mesh",1);
  Oxs_DepKey<Oxs_Ext> key
    = director->FindExtObject((*FindInitValue("mesh"))[0]);
  Oxs_Mesh* mesh = dynamic_cast<Oxs_Mesh*>(key.GetPtr());
  if(mesh==NULL) {
    string msg="No Oxs_Mesh object found with name "
      + (*FindInitValue("mesh"))[0];
    throw Oxs_Ext::Error(msg.c_str());
  }
  meshkey.Set(mesh);
  DeleteInitValue("mesh");

  VerifyAllInitArgsUsed();

  // Fill member variable vec with values from import Tcl script.
  size=mesh->Size();
  vec.AdjustSize(mesh);

  char extents[1024];
  Oxs_Box bbox;
  mesh->GetBoundingBox(bbox);
  Oc_Snprintf(extents,sizeof(extents),
              "%.16g %.16g %.16g %.16g %.16g %.16g",
              bbox.GetMinX(),bbox.GetMinY(),bbox.GetMinZ(),
              bbox.GetMaxX(),bbox.GetMaxY(),bbox.GetMaxZ());

  char cmdbuf[4096];
  ThreeVector loc;
  Oxs_SplitList veclist;

  for(UINT4m i=0;i<size;i++) {
    mesh->Center(i,loc);
    Oc_Snprintf(cmdbuf,sizeof(cmdbuf),
                "%s %.16g %.16g %.16g %s",
                script.c_str(),loc.x,loc.y,loc.z,extents);
    if(Tcl_Eval(safe_interp,cmdbuf)!=TCL_OK) {
      string msg=string("Error detected during construction of ")
        + string(InstanceName()) + string("; Tcl script error: ")
        + string(Tcl_GetStringResult(safe_interp));
      throw Oxs_Ext::Error(msg.c_str());
    }
    veclist.Split(Tcl_GetStringResult(safe_interp));
    if(veclist.Count()!=3) {
      string msg
        = string("Error detected during construction of ")
        + string(InstanceName())
        + string("Return script value is not a triplet of 3 numbers: ")
        + string(Tcl_GetStringResult(safe_interp));
      throw Oxs_Ext::Error(msg.c_str());
    }
    vec[i].x=Nb_Atof(veclist[0]);
    vec[i].y=Nb_Atof(veclist[1]);
    vec[i].z=Nb_Atof(veclist[2]);
    if(setnorm) vec[i].SetMag(norm);
  }
}

BOOL Oxs_TSFVectorField::Init()
{ return 1; }

Oxs_TSFVectorField::~Oxs_TSFVectorField()
{}

ThreeVector
Oxs_TSFVectorField::operator[](UINT4m index) const
{
  if(index>=size) {
    char buf[512];
    Oc_Snprintf(buf,sizeof(buf),
		"In object %s, "
		"Oxs_TSFVectorField::operator[] const: "
		"Array out-of-bounds; index=%u, size=%u",
		InstanceName(),index,size);
    throw Oxs_Ext::Error(buf);
  }
  return vec[index];
}


