/* FILE: ucexchange.cc            -*-Mode: c++-*-
 *
 * Uniform exchange energy on cubic mesh, derived from Oxs_Energy class.
 *
 */

#include "nb.h"
#include "depkey.h"
#include "director.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "cubicmesh.h"
#include "ucexchange.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_UCExchange);

/* End includes */


// Constructor
Oxs_UCExchange::Oxs_UCExchange(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  CheckInitValueParamCount("Ms",1);
  Ms=Nb_Atof((*FindInitValue("Ms"))[0].c_str());
  DeleteInitValue("Ms");

  CheckInitValueParamCount("A",1);
  A=Nb_Atof((*FindInitValue("A"))[0].c_str());
  DeleteInitValue("A");

  CheckInitValueParamCount("mesh",1);
  const string& meshname = (*FindInitValue("mesh"))[0];
  Oxs_DepKey<Oxs_Ext> tempkey = director->FindExtObject(meshname);
  if(tempkey.GetPtr()==NULL) {
    string msg="No Oxs_Mesh object found with name " + meshname;
    throw Oxs_Ext::Error(msg.c_str());
  }
  Oxs_CubicMesh* mesh = dynamic_cast<Oxs_CubicMesh*>(tempkey.GetPtr());
  if(mesh==NULL) {
    string msg="Object " + meshname + " is not a cubic mesh.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  meshkey.Set(mesh); // Dependency lock is held until *this is deleted.
  DeleteInitValue("mesh");

  VerifyAllInitArgsUsed();
}

Oxs_UCExchange::~Oxs_UCExchange()
{}

BOOL Oxs_UCExchange::Init()
{
  return 1;
}

void Oxs_UCExchange::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  // Check that the mesh hasn't changed.
  const Oxs_CubicMesh* mesh = meshkey.GetPtr();
  if(mesh != state.mesh) {
    string msg="Input mesh to Oxs_UCExchange::GetEnergyAndField()"
      " routine of object " +  string(InstanceName())
      + " is different than the one used to initialsize this object.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  const Oxs_MeshValue<ThreeVector>& spin = state.spin;

  UINT4m xdim = mesh->XDim();
  UINT4m ydim = mesh->YDim();
  UINT4m zdim = mesh->ZDim();
  UINT4m xydim = xdim*ydim;

  REAL8m cellsize = mesh->EdgeLength();
  REAL8m ecoef = -A/(cellsize*cellsize);
  REAL8m hcoef = 2*A/(MU0*Ms*cellsize*cellsize);
  for(UINT4m z=0;z<zdim;z++) {
    for(UINT4m y=0;y<ydim;y++) {
      for(UINT4m x=0;x<xdim;x++) {
	UINT4m i = mesh->Index(x,y,z); // Get base linear address
	ThreeVector base = spin[i];
	ThreeVector sum(0.,0.,0.);
	if(z>0)      sum += (spin[i - xydim] - base);
	if(y>0)      sum += (spin[i - xdim]  - base);
	if(x>0)      sum += (spin[i - 1]     - base);
	if(x<xdim-1) sum += (spin[i + 1]     - base);
	if(y<ydim-1) sum += (spin[i + xdim]  - base);
	if(z<zdim-1) sum += (spin[i + xydim] - base);
	field[i] = hcoef * sum;
	energy[i] = ecoef * (sum * base);
      }
    }
  }
}




