/* FILE: uuanisotropy.cc            -*-Mode: c++-*-
 *
 * Uniform Uniaxial Anisotropy, derived from Oxs_Energy class.
 *
 */

#include "oc.h"
#include "nb.h"
#include "threevector.h"
#include "director.h"
#include "simstate.h"
#include "ext.h"
#include "depkey.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "cubicmesh.h"
#include "uuanisotropy.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_UUAnisotropy);

/* End includes */


// Constructor
Oxs_UUAnisotropy::Oxs_UUAnisotropy(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  CheckInitValueParamCount("K1",1);
  K1=Nb_Atof((*FindInitValue("K1"))[0].c_str());
  DeleteInitValue("K1");

  CheckInitValueParamCount("Ms",1);
  Ms=Nb_Atof((*FindInitValue("Ms"))[0].c_str());
  DeleteInitValue("Ms");

  CheckInitValueParamCount("axis",3);
  axis.x=Nb_Atof((*FindInitValue("axis"))[0].c_str());
  axis.y=Nb_Atof((*FindInitValue("axis"))[1].c_str());
  axis.z=Nb_Atof((*FindInitValue("axis"))[2].c_str());
  DeleteInitValue("axis");

  VerifyAllInitArgsUsed();

  if(Ms==0.0) {
    K1=0.0; Ms=1.0; // Safety
  }

  REAL8m magsq=axis.MagSq();
  if(magsq==0.0) {
    string msg="Invalid MIF input block detected for object "
      + string(InstanceName())
      + ": Specified anisotropy axis is (0,0,0)";
    throw Oxs_Ext::Error(msg.c_str());
  }
  if(magsq!=1.0) axis *= 1.0/sqrt(magsq);
}

BOOL Oxs_UUAnisotropy::Init()
{ return 1; }

void Oxs_UUAnisotropy::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  UINT4m size = state.mesh->Size();

  REAL8m energy_mult = -K1;
  REAL8m field_mult = 2*K1/(Ms*MU0);
  if(K1>0) {
    for(UINT4m i=0;i<size;++i) {
      REAL8m temp = axis*state.spin[i];
      energy[i] = energy_mult*(temp*temp-1.0); // Make easy axis zero energy
      field[i] = (field_mult*temp)*axis;
    }
  } else {
    for(UINT4m i=0;i<size;++i) {
      REAL8m temp = axis*state.spin[i];
      energy[i] = energy_mult*temp*temp; // Easy plane is zero energy
      field[i] = (field_mult*temp)*axis;
    }
  }
}
