# FILE: mif.tcl
#
# Class for holding micromagnetic problem spec
#
# The input file should look like
#
#       # MIF 2.0
#       Specify Energy:UniformExchange: {
#           A 21e-12
#       }
#       Specify Energy:UniformAnisotropy: {
#           Type Uniaxial
#           K1   0.5e3
#           Axis 1 0 0
#       }
#
# This is evaluated inside a safe interpreter, so "unsafe" commands
# are not available
#
Oc_Class Oxs_Mif {
    private variable speckeys   ;# Ordered list of keys
    private array variable spec
    public variable safe_interp = {}  ;# Slave safe interpreter used to
    ;## read input mif file.  A new interpreter is created each time a
    ;## mif file is read.

    Constructor { args } {
        eval $this Configure $args
        set speckeys {}
    }

    method SetSpec { key args } {
	set ac [llength $args]
	if {$ac != 1} {
	    return -code error \
		    "Bad argument count ($ac) in Specify block (KEY $key)"
	}
	if {[string first ":" $key]<0} {
	    # Append default instance name (empty string)
	    append key ":"
	}
	if {[info exists spec($key)]} {
	    return -code error "Key $key already in use"
	}
        lappend speckeys $key
        set spec($key) [lindex $args 0]
    }

    method UnsetSpec { key } {
	# If $key is {}, then clears out all specs.  This does not
        # raise an error even if there are no specs.
        #   If $key is not {}, then removes only that spec, from
        # both the spec array and the speckeys list.  This operation
        # will raise an error if the spec $key is not set.
        if {[string match {} $key]} {
            catch {unset spec}
            set speckeys {}
        } else {
            unset spec($key)
            set keyindex [lsearch -exact $speckeys $key]
            if {$keyindex>=0} {
                set speckeys [lreplace $speckeys $keyindex $keyindex]
            }
        }
    }
    # It might also be useful to have a 'ReplaceSpec' method
    # that does an Unset + Set without changing key order.

    method Read { filename } {
        # Fill spec from filename.  A fresh safe interpreter is created
        # each time this function is called.  Between calls to Read,
        # the safe interpreter is held in member variable safe_interp,
        # where it may be accessed by MIF element constructors.

	# Empty out any previous specifications (should we do this?)
	$this UnsetSpec {}
        if {![string match {} $safe_interp]} {
            interp delete $safe_interp
	    set safe_interp {}
        }

	# Read file into a string.  Hopefully this isn't too big.
	set chan [open $filename]
	set filestr [read $chan]
	close $chan

	# Check that first line matches "# MIF 2.0"
	if {![regexp -- "^#\[ |\t\]*MIF\[ |\t\]*2.0\[ |\t\]*\n" \
		$filestr]} {
	    return -code error \
		    "Input file $filename not in MIF 2.0 format"
	}

	# Setup safe interpreter to source filestr
	set safe_interp [interp create -safe]
        interp alias $safe_interp Specify {} $this SetSpec
	interp alias $safe_interp Clear {} $this UnsetSpec
        set errcode [catch {$safe_interp eval $filestr} errmsg]
	if {$errcode} {
	    global errorInfo errorCode
	    foreach {ei ec} [list $errorInfo $errorCode] {break}
	    # Error occurred sourcing $filename
	    $this UnsetSpec {}   ;# Clear out partial problem spec
	    return -code error -errorinfo $ei -errorcode $ec \
		    "Error processing file $filename: $errmsg"
	}
    }

    method GetKeys {} {
        return $speckeys
    }

    method GetValue { name } {
        return $spec($name)
    }

    method Dump {} {
        # Return entire spec, in order, in MIF 2.0 file format.
        set specstr "# MIF 2.0\n"
        foreach k $speckeys {
            append specstr "[list Specify $k $spec($k)]\n"
        }
        return $specstr
    }

    Destructor {
        if {![string match {} $safe_interp]} {
            interp delete $safe_interp
        }
    }
}
