# FILE: solver.tcl
#
# Class for controlling micromagnetic solver
#

Oc_Class Oxs_Solver {
    const private variable mif

    private variable solver_state = Invalid
    ## Should be one of Pause, OneStep, Relax, Run, or Init.
    ## Set by calls to ChangeSolverState.  The 'after idle'
    ## method BackgroundStep checks the value of this variable at
    ## the start and end of processing.  Init the initial value
    ## set during initialization and reset operations, and should
    ## not be otherwise used.  This value can be read by any method
    ## but should only be changed by the ChangeSolverState method.

    private variable field_index = 0

    private variable background_id = {}

    # Events generated by this class:
    #     Iteration    --- ODE step complete
    #     ControlPoint --- Control point reached (at current field)
    #     ChangeState  --- Solver changing state (%state is new state)
    #     RunDone      --- New problem, reset, solver closing, or FieldStep
    #                      attempted with no more steps in list
    #     Delete       --- Instance being deleted

    # Note: A given ODE iteration my raise multiple events, but
    #       CheckPoint events are not raised more than once on the same
    #       iteration.

    const private common reschedule_delay

    ClassConstructor {
        global tcl_platform
        set reschedule_delay 0
        # Socket reading on Windows needs an opening in the
        # event scheduling to work properly.  On other platforms
        # reschedule as fast as possible for maximum compute
        # performance.
        if {[string match windows $tcl_platform(platform)]} {
            set reschedule_delay 1
        }
    }
    Constructor { args } {
        eval $this Configure $args
	Oxs_Mif New mif
    }
    method ProbInit { mif_name {restart_flag 0}} {
        if {[string match {} $mif_name]} {
            error "Invalid mif object passed to Oxs_Solver::ProbInit"
        }
        $this Reset $mif_name $restart_flag
    }
    method Reset { {mif_name {}} {restart_flag 0} } {
        $this ChangeSolverState Init
        if {![string match {} $mif_name]} {
            $mif Read $mif_name
            Oxs_ProbInit $mif
        } else {
            Oxs_ProbReset
        }
	$this ChangeSolverState Pause
    }

    private method FieldChangeRequest { new_field_index } \
            { field_index } {
        # Confer with director abount field change request
        # Returns 1 if field request honored, 0 otherwise
        if {$new_field_index<0} {
            return 0
        }
        set field_index $new_field_index
        return 1
    }

    method FieldStepRequest { offset } { field_index } {
        return [$this FieldChangeRequest [expr $field_index + $offset]]
    }

    method ControlPointCheck {} {
        # Returns 1 if stopping criteria are met, 0 otherwise.
        return 1 ;###
    }

    ;### NOTE: The following method should be private, but is run via
    ;###   the update stack so must be technically public.
    method BackgroundStep { stop_check } \
            {background_id solver_state field_index reschedule_delay} {
        # This method should be called via 'after idle'.  At
        # the start of processing, this method first checks the
        # value of solver_state, and returns immediately if
        # solver_state==Pause.  Otherwise, if stop_check is false,
        # then one step is made in the ODE, w/o any update calls.
        # After this, the stop conditions are checked, and the
        # method may reschedule itself, depending on solver_state
        # and the result of ControlPointCheck.  The rescheduling
        # occurs the same way, independent of the value of stop_check,
        # so a Relax or a Run sequence can be initiated by calling
        # this routine with stop_check set to 1 to disable unnecessary
        # ODE stepping at the current field.

        if {[string match {} $background_id]} {
            # Safety check
            error "solver method BackgroundStep call scheduling error"
        }

        set background_id {}
        if {[string match Pause $solver_state]} { return }

        # Do some work
        if {!$stop_check} { Oxs_Run 1 }

        # Check control_point and field_step requirements, and generate
        # events and callbacks as appropriate.
        set control_point_event 0
        set fieldstep_attempt 0
        if {[$this ControlPointCheck]} {
            set control_point_event 1
            if {[string match Run $solver_state]} {
                set fieldstep_attempt 1
            }
        }
        if {$control_point_event} {
            # Control point events get called with old field values
            # in place.
            if {[catch {Oc_EventHandler Generate $this ControlPoint} \
                    msg]} {
                bgerror $msg
            }
        }
        set fieldstep_event 0
        if {$fieldstep_attempt} {
            set fieldstep_event [$this FieldStepRequest 1]
            if {!$fieldstep_event} {
                if {[catch {Oc_EventHandler Generate $this RunDone} \
                        msg]} {
                    bgerror $msg
                }
            }
        }

        # Reschedule if necessary, otherwise set solver_state to Pause.
        # (Also do safety check on $solver_state.)
        # NB: 'Pause' is a possible state, because the Oc_EventHandlers
        #     may modify the solver state.
        set reschedule 0
        switch -exact -- $solver_state {
            Run     {
                if {!$control_point_event || $fieldstep_event} {
                    set reschedule 1
                }
            }
            Relax   {
                if {!$control_point_event} {
                    set reschedule 1
                }
            }
            Pause  -
            OneStep {}
            default {
                error "Invalid solver_state detected in\
                        BackgroundStep: ->${solver_state}<-"
            }
        }
        if {$reschedule} {
	    set new_stop_check $fieldstep_event
            set background_id [after $reschedule_delay \
                    set [$this GlobalName background_id] \
                    \[after idle \
		    [list $this BackgroundStep $new_stop_check]\]]
            ## We need to schedule this from the non-idletask stack,
            ## because if Tcl_DoOneEvent processes *any* idletasks, it
            ## processes *all*, including apparently ones that go into
            ## the list after idletask processing has commenced.  The
            ## above seems to work, but Dean Brettle <brettle@his.com>
            ## claims that there is a race condition in this as well
            ## (see his comp.lang.tcl posting on 1998/02/15).  I haven't
            ## looked at the Tcl core source code, but it may be that
            ## a Tcl 'update' call with the above in place may be evil
            ## as well.  If a problem is observed with the above, it is
            ## likely that changing the 'after 0' to, say, 'after 20' will
            ## better hide the problem.  -mjd, 1998/04/07.
            ##
            ## See also the ClassConstructor - DGP 1998/11/04.
        } else {
            $this ChangeSolverState Pause
        }
    }

    method ChangeSolverState { newstate } {
        # 'newstate' should be one of OneStep, Relax, Run
        #  or Pause.
        if {[string match $solver_state $newstate]} {return} ;# No change

        # Rather than relying on the state of background_id, play
        # it safe and always de-schedule background_id, and
        # re-schedule if necessary.  (Presumably ChangeSolverState
        # won't be called inside a tight loop!)
        if {![string match {} $background_id]} {
            after cancel $background_id
            set background_id {}
        }

        switch -exact -- $newstate {
            OneStep {
                set background_id \
                        [after idle [list $this BackgroundStep 0]]
            }
            Relax   -
            Run    {
                set background_id \
                        [after idle [list $this BackgroundStep 1]]
            }
            Init {}
            Pause {}
            default {
                error "Invalid solver_state request to\
                        ChangeSolverState: $newstate"
            }
        }
        set solver_state $newstate
	if {[catch {
            Oc_EventHandler Generate $this ChangeState \
                    -state $solver_state
	} msg]} {
	    bgerror $msg
	}
    }
    method GetODEStepCount {} {
        return 0
    }
    method GetFieldStepCount {} { field_index } {
        return $field_index
    }
    method GetTabularData {} {} {
        return {}
    }
    method WriteFieldFile { args } {} {
        return [list {} {}]
    }
    method WriteMagFile { args } {} {
        return [list {} {}]
    }
    Destructor {
        $this ChangeSolverState Pause
        ## In particular, clear "after" stack

        Oc_EventHandler Generate $this RunDone
        Oc_EventHandler Generate $this Delete
        Oc_EventHandler DeleteGroup $this
        if {[info exists mif]} {
            catch { $mif Delete }
        }
    }
}
