# FILE: oxsi.tcl
#
# Interactive interface to the oxs shell.

# Support libraries
package require Oc 1.1
package require Oxs 1.2
package require Net

Oc_IgnoreTermLoss  ;# Try to keep going, even if controlling terminal
## goes down.

Oc_Main SetAppName Oxsi
Oc_Main SetVersion 1.2.0.0
regexp \\\044Date:(.*)\\\044 {$Date: 2000/08/01 22:32:18 $} _ date
Oc_Main SetDate [string trim $date]
Oc_Main SetAuthor [Oc_Person Lookup mjd]
Oc_Main SetHelpURL [Oc_Url FromFilename [file join [file dirname \
        [file dirname [file dirname [Oc_DirectPathname [info \
        script]]]]] doc userguide 3D_Micromagnetic_Solver.html]]

Oc_CommandLine Option restart {
	{flag {expr {![catch {expr {$flag && $flag}}]}} {= 0 or 1}}
    } {
	global restart_option; set restart_option $flag
} {1 => use <basename>.log file to restart simulation}
set restart_option 0

Oc_CommandLine Parse $argv

if {[Oc_Main HasTk]} {
    wm withdraw .        ;# "." is just an empty window.
    package require Ow
    Ow_SetIcon .
} else {
    # Set up to write log messages to oommf/mmsolve2d.errors.
    # Note that the solver class will later re-direct this to its own log
    # file.
    FileLogger SetFile \
	    [file join [Oc_Main GetOOMMFRootDir] oxsi.errors]
    Oc_Log SetLogHandler [list FileLogger Log] panic Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] error Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] warning Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] info Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] panic
    Oc_Log SetLogHandler [list FileLogger Log] error
    Oc_Log SetLogHandler [list FileLogger Log] warning
    Oc_Log SetLogHandler [list FileLogger Log] info
}

set commands {
    Solver	{{LoadProblem RemoteRelayInput ProblemDescription}}
}

array set inputs {
    ProblemDescription	{ProbEd
			 GetProbDescFile
			 LoadProbDescFile
                        }
}
array set events {}
array set outputs {}

Oc_EventHandler Bindtags LoadProbDescFile LoadProbDescFile
proc LoadProbDescFile { mif_file } {
    global commands events inputs outputs
#    catch { unset inputs } ;###
    catch { unset outputs }  ;# "outputs" is _not_ always set!
    catch { unset events }
    catch { unset commands } 
#    array set inputs {}   ;###
# The commands which can be invoked remotely
set commands {
    Solver	{
		 {Reset ResetThread} 
                 {LoadProblem RemoteRelayInput ProblemDescription}
                 {Run ChangeSolverState Run}  
                 {Relax ChangeSolverState Relax}  
                 {Pause ChangeSolverState Pause}  
		 {Field- FieldAdjust -1}
		 {Field+ FieldAdjust 1}
		}
}

# The events generated by this thread.
array set events {
    Iteration	 {1 solver GetODEStepCount}
    ControlPoint {1 solver GetFieldStepCount}
}

# The quantities which this thread produces as output
#	The last entry should be moved to be set by controller
array set outputs {
    Magnetization	{vectorField
			 {Iteration ControlPoint}
			 solver
			 WriteMagFile
			 datafile
                        }
    TotalField 		{vectorField
			 {Iteration ControlPoint}
			 solver
			 WriteFieldFile
			 datafile
                        }
    DataTable		{DataTable
			 {Iteration ControlPoint}
			 ""
			 MakeDataTable
			 DataTable
			}
}


    global solver protocol restart_option
    if {![info exists solver]} {
	# First pass---new solver
	Oxs_Solver New solver
	Oc_EventHandler Generate LoadProbDescFile InterfaceChange
	# Examine this:
	Oc_EventHandler New _ $solver Delete [list $protocol Delete] \
                -groups [list $solver $protocol]
	Oc_EventHandler New _ $solver RunDone [list RunEndNotify]
        Oc_EventHandler New _ Oc_Main Exit [list $solver Delete] \
                -groups [list $solver]
    }
    $solver ProbInit $mif_file $restart_option
    file delete $mif_file
}

Net_Protocol New protocol -name "OOMMF solver protocol 0.0"
$protocol AddMessage start HasGui {} {
    global solver
    if {![info exists solver]} {
	Oc_EventHandler New _ LoadProbDescFile InterfaceChange \
		[list $connection Tell InterfaceChange] -oneshot 1 \
		-groups [list $connection]
    }
    Oc_EventHandler New _ oxs_solver ChangeState \
	    [list $connection Tell Status %state] \
	    -groups [list $connection]
    return [list start [list 0 1]]
}
$protocol AddMessage start Commands {} {
    global commands
    return [list start [list 0 $commands]]
}
$protocol AddMessage start Outputs {} {
    global outputs
    return [list start [list 0 [array get outputs]]]
}
$protocol AddMessage start Inputs {} {
    global inputs
    return [list start [list 0 [array get inputs]]]
}
$protocol AddMessage start Events {} {
    global events
    return [list start [list 0 [array get events]]]
}
$protocol AddMessage start SetHandler {args} {
    return [list start [list 0 [eval SetHandler $args]]]
}
$protocol AddMessage start SetInput {args} {
    return [list start [list 0 [eval SetInput $args]]]
}
$protocol AddMessage start UnsetHandler {args} {
    return [list start [list 0 [eval UnsetHandler $args]]]
}
$protocol AddMessage start InteractiveOutput {args} {
    return [list start [list 0 [eval InteractiveOutput $args]]]
}
$protocol AddMessage start RemoteRelayInput {args} {
    return [list start [list 0 [eval RemoteRelayInput $args]]]
}
$protocol AddMessage start FieldAdjust {args} {
    return [list start [list 0 [eval FieldAdjust $args]]]
}
$protocol AddMessage start ResetThread {args} {
    return [list start [list 0 [eval ResetThread $args]]]
}
$protocol AddMessage start ChangeSolverState {args} {
    return [list start [list 0 [eval ChangeSolverState $args]]]
}
$protocol AddMessage start delete {} {
    global protocol
    return [list start [list 0 [after 100 $protocol Delete]]]
}
Oc_EventHandler New _ $protocol Delete [list after 2000 exit] -oneshot 1
Net_Server New server -protocol $protocol -alias Oxsi
$server Start 0

# Delay "nice" of this process until any children are spawned.
Net_Account New a
if {[$a Ready]} {
    Oc_MakeNice
} else {   
    Oc_EventHandler New _ $a Ready Oc_MakeNice -oneshot 1
}

proc SetInput {data host acct pid} {
    Oc_EventHandler DeleteGroup SetInput
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {}
    Oc_EventHandler New _ RemoteRelayInput $data \
	    [list RelayInput $data $t] -groups [list $t SetInput]
}

proc SetHandler {event data host acct pid cnt} {
    global events
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {}
    Oc_EventHandler DeleteGroup $event-$data-$t
    if {[string match interactive $event]} {
	Oc_EventHandler New _ InteractiveOutput $data \
		[list RelayOutput $data $t] \
		-groups [list $t $event-$data-$t]
    } else {
	set evinfo $events($event)
	set srcName [lindex $evinfo 1]
	upvar #0 $srcName src
	Oc_EventHandler New _ $src $event \
		[list GenericHandler $event $data $cnt $t] \
		-groups [list $t $event-$data-$t]
    }
    global threadrefs
    set threadrefs($event-$data-$t) 1
}

proc UnsetHandler {event data host acct pid} {
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {}
    Oc_EventHandler DeleteGroup $event-$data-$t
    global threadrefs
    catch {unset threadrefs($event-$data-$t)}
    if {![llength [array names threadrefs *-$t]]} {
	$t Delete
    }
}

proc GenericHandler {event data cnt thread} {
    global events
    set evinfo $events($event)
    if {[lindex $evinfo 0]} {
        set srcName [lindex $evinfo 1]
        upvar #0 $srcName src
        set evcount [eval $src [lindex $evinfo 2]]
        if {$evcount % $cnt} {
            return
        }
    }
    RelayOutput $data $thread
}

Oc_EventHandler Bindtags InteractiveOutput InteractiveOutput
proc InteractiveOutput { data } {
    Oc_EventHandler Generate InteractiveOutput $data
}

# Send the data indicated by 'data' to the process represented by 'thread'
proc RelayOutput { data thread } {
    # Check to see that 'thread' is ready to send commands to its process
    if {[catch {$thread Ready} ready] || !$ready} {
        # 'thread' isn't ready.  Two possible reasons:
        if {![string match {} [info commands $thread]]} {
            # 1. Thread exists, but just isn't ready yet.
            #    --> Catch its Ready event
            Oc_EventHandler New _ $thread Ready \
                    [list RelayOutput $data $thread] \
                    -oneshot 1 -groups [list $thread]
	}
        # 2. There is no such thread instance --> give up.
    } else {
        # Thread is ready.  Construct and send data.
        global outputs
        set opinfo $outputs($data)
        set srcName [lindex $opinfo 2]
        if {[string length $srcName]} {
            upvar #0 $srcName src
            set datavalue [eval $src [lindex $opinfo 3]]
        } else {
            set datavalue [eval [lindex $opinfo 3]]
        }
        set method [lindex $opinfo 4]
        $thread Send $method $datavalue
    }
}

Oc_EventHandler Bindtags RemoteRelayInput RemoteRelayInput
proc RemoteRelayInput { data } {
    Oc_EventHandler Generate RemoteRelayInput $data
}

# Send request for 'data' to the process represented by 'thread'
proc RelayInput { data thread } {
    # Check to see that 'thread' is ready to send commands to its process
    if {[catch {$thread Ready} ready] || !$ready} {
        # 'thread' isn't ready.  Two possible reasons:
        if {![string match {} [info commands $thread]]} {
            # 1. Thread exists, but just isn't ready yet.
            #    --> Catch its Ready event
            Oc_EventHandler New _ $thread Ready \
                    [list RelayInput $data $thread] \
                    -oneshot 1 -groups [list $thread]
        }
        # 2. There is no such thread instance --> give up.
    } else {
        # Thread is ready.  Construct and send request.
        global inputs 
        set ininfo $inputs($data)
        set qid [$thread Send [lindex $ininfo 1]]    
        Oc_EventHandler New _ $thread Reply$qid \
               [list ReceiveRelayInputReply $thread $data] \
               -oneshot 1 -groups [list $thread]
        # Consider handler for timeout event?
    }
}

proc ReceiveRelayInputReply {thread data} {
    global inputs protocol
    set reply [$thread Get]
    if {![lindex $reply 0]} {
        set ininfo $inputs($data)
        eval [lindex $ininfo 2] [lrange $reply 1 end]
    }
}

proc MakeDataTable {} {
    global solver
    return [$solver GetTabularData]
}

proc FieldAdjust { step } {
    global solver 
    if {$step != 0} {
	$solver FieldStepRequest $step
    }
}

proc ResetThread { } {
    global solver
    $solver Reset
}

proc ChangeSolverState { newstate } {
    global solver
    $solver ChangeSolverState $newstate
}

proc RunEndNotify {} {
    # Send end-of-run message to each thread supporting
    # the DataTable protocol
    set eormsg [list [list "@Filename:" {} 0]]
    foreach thread [Net_Thread Instances] {
	if {[string match DataTable [$thread Protocol]]} {
	    $thread Send DataTable $eormsg
	}
    }
}

vwait forever

