/* FILE: depkey.h                 -*-Mode: c++-*-
 *
 * Dependency key class, intended for use with the Oxs_DepLock family.
 *
 */

#ifndef _OXS_DEPKEY
#define _OXS_DEPKEY

#include "oc.h"
#include "oxsexcept.h"  // Needed in template definitions

/* End includes */

////////////////////////////////////////////////////////////////////////
// Key class.  Class T should be a child of Oxs_DepLock.
template<class T> class Oxs_DepKey
{
private:
  T* ptr; // A dependency lock is held iff ptr!=NULL
public:
  void Release(); // Release lock and set ptr to NULL

  void Set(T* ptr_); // Sets ptr to ptr_, releases any current
  /// dependency lock, and if ptr_!=NULL, registers dependency

  Oxs_DepKey() : ptr(0) {}
  Oxs_DepKey(T* ptr_) : ptr(0) { Set(ptr_); }
  ~Oxs_DepKey() { Release(); }

  // Copy constructor.  Duplicates ptr and registers dependency
  Oxs_DepKey(const Oxs_DepKey<T>& other);

  Oxs_DepKey<T>& operator=(const Oxs_DepKey<T>& other);
  /// Same behavior as copy constructor

  void Swap(Oxs_DepKey<T>& other)
  { T* tptr=ptr; ptr=other.ptr; other.ptr=tptr; }

  T* GetPtr() { return ptr; }
  const T* GetPtr() const { return ptr; }
};

////////////////////////////////////////////////////////////////////////
// Include all definitions here.  This is required by compiler/linkers
// using what gcc calls the "Borland model," as opposed to the "Cfront
// model."  The latter is distinguished by the use of template
// repositories.  With the Cfront model, template definitions can be
// put in a file (*.cc) separate from the header file (*.h).  Putting
// the definitions in the header file along with declarations slows down
// compilation, because the definitions are re-compiled in each including
// file, but appears to be more portable.

template<class T>
void Oxs_DepKey<T>::Release()
{
  if(ptr!=NULL) {
    ptr->ReleaseDependency();
    ptr=NULL;
  }
}

template<class T>
void Oxs_DepKey<T>::Set(T* tptr)
{
  Release();
  if(tptr!=NULL) {
    if(!tptr->SetDependency())
      OC_THROW(Oxs_BadLock,"Unable to register dependency");
  }
  ptr=tptr;
}

// Copy constructor.  Duplicates ptr and registers dependency
template<class T>
Oxs_DepKey<T>::Oxs_DepKey(const Oxs_DepKey<T>& other)
{
  if(other.ptr==NULL) {
    ptr=NULL;
  } else {
    if(!other.ptr->SetDependency())
      OC_THROW(Oxs_BadLock,"Unable to register dependency");
    ptr=other.ptr;
  }
}

template<class T>
Oxs_DepKey<T>& Oxs_DepKey<T>::operator=(const Oxs_DepKey<T>& other)
{ // Same behavior as copy constructor
  if(this==&other) return *this; // Nothing to do
  Release();
  if(other.ptr==NULL) {
    ptr=NULL;
  } else {
    if(!other.ptr->SetDependency())
      OC_THROW(Oxs_BadLock,"Unable to register dependency");
    ptr=other.ptr;
  }
  return *this;
}

#endif // _OXS_DEPKEY
