/* FILE: director.cc                 -*-Mode: c++-*-
 *
 * 3D solver director class.
 *
 */

#include <vector>
#include <string>
#include "oc.h"

#include "director.h"
#include "simstate.h"
#include "ext.h"
#include "energy.h"
#include "util.h"
#include "depkey.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "threevector.h"

/* End includes */

int myEval(Tcl_Interp* interp,const string& cmd)
{
  char* buf=new char[cmd.length()+1];
  strcpy(buf,cmd.c_str());
  int errcode=Tcl_Eval(interp,buf);
  delete[] buf;
  return errcode;
}

// The following two routines make sure that all Oxs_Director
// instances are destroyed as part of a Tcl_Exit().
void Oxs_Director::ExitProc(ClientData clientData)
{
    Oxs_Director *d = (Oxs_Director *) clientData;
    delete d;
}

Oxs_Director::Oxs_Director(Tcl_Interp *i) 
  : problem_loaded(0),interp(i), simulation_state_reserve_count(0)
{
  Tcl_CreateExitHandler((Tcl_ExitProc *) ExitProc, (ClientData) this);
}

int Oxs_Director::ProbReset()
{ // Sets micromag problem to initial state
  // Run Init() on all Oxs_Ext's
  vector<Oxs_Ext*>::iterator it = ext_obj.begin();
  while(it!=ext_obj.end()) {
    if(!((*it)->Init())) return TCL_ERROR;
    it++;
  }
  return TCL_OK;
}

int Oxs_Director::ProbInit(const char* mif)
{
  int i,error_code;
  string cmd;

  Release();
  simulation_state_reserve_count=0;  // Safety

  // Get pointer to safe interpreter used by mif object
  cmd = mif;
  cmd.append(" Cget -safe_interp");
  error_code=myEval(interp,cmd);
  if(error_code!=TCL_OK) return error_code;
  Oc_AutoBuf slave_name(Tcl_GetStringResult(interp));
  Tcl_Interp* safe_interp = Tcl_GetSlave(interp,slave_name);
  if(safe_interp==NULL || *((char *)slave_name)=='\0') {
    Tcl_AppendResult(interp,
                     "Safe interpreter not set in Oxs_Mif object ",
		     mif,(char *)NULL);
    return TCL_ERROR;
  }

  // Get list of spec blocks in mif object
  cmd = mif;
  cmd.append(" GetKeys");
  error_code=myEval(interp,cmd);
  if(error_code!=TCL_OK) return error_code;

  Oxs_SplitList arglist;
  error_code = arglist.Split(Tcl_GetStringResult(interp));
  if(error_code!=TCL_OK) return error_code;

  // Create Oxs_Ext object for each spec block, storing pointer
  // in ext_obj, and mapping on id in ext_map.
  string bad_specs;
  for(i=0;i<arglist.Count();i++) {
    cmd=mif;
    cmd.append(" GetValue ");
    cmd.append(arglist[i]);
    error_code=myEval(interp,cmd);
    if(error_code!=TCL_OK) return error_code;
    string argstr = Tcl_GetStringResult(interp);

    Oxs_Ext* spec_ptr=NULL;
    try {
      spec_ptr=Oxs_Ext::MakeNew(arglist[i],this,
				safe_interp,argstr.c_str());
    }
    catch (Oxs_Ext::Error& err) {
      Release();  // Delete any Oxs_Ext objects already constructed
      Tcl_AppendResult(interp,"Oxs_Ext initialization error---",
		       err.msg,(char *)NULL);
      return TCL_ERROR;
    }

    if(spec_ptr!=NULL) {
      string idname=spec_ptr->InstanceName();
      if(ext_map.find(idname)!=ext_map.end()) {
	// Name already in use.  Append to bad spec list and delete.
	delete spec_ptr;
	bad_specs += string("\n ") + string(arglist[i])
	  + string(" (ID name ") + idname + string("already in use.)");
      } else {
	ext_map[idname]=spec_ptr;  // Store mapping
	// Store in ordered list
	ext_obj.push_back(spec_ptr);
	// If this is an Oxs_Energy object, then append it to
	// the energy_obj list as well.
	Oxs_Energy* energy_ptr
	  = dynamic_cast<Oxs_Energy*>(spec_ptr);
	if(energy_ptr!=NULL)
	  energy_obj.push_back(energy_ptr);
      }
    } else {
      bad_specs += string("\n  ") + string(arglist[i]);
    }
  }

  // Report bad spec blocks
  Tcl_ResetResult(interp);
  if(bad_specs.length()>0) {
    Release();  // Delete all Oxs_Ext objects
    Tcl_AppendResult(interp,"Bad Spec Blocks---",
		     bad_specs.c_str(),(char *)NULL);
    return TCL_ERROR;
  }

  // Run Init() on all Oxs_Ext's
  ProbReset();
  problem_loaded=1;

  return TCL_OK;
}

void Oxs_Director::Release()
{
  problem_loaded=0;

  // Release director scratch space
  temp_energy.Release();
  temp_field.Release();

  // Release all Oxs_Ext's, from the back forward.  The
  // order is important so that objects at the back of the
  // list can release any dependency locks held on objects
  // at the front of the list.
  vector<Oxs_Ext*>::reverse_iterator it = ext_obj.rbegin();
  while(it!=ext_obj.rend()) {
    delete (*it);
    ++it;
  }
  // Release *all* pointers to these objects.
  ext_obj.clear();
  ext_map.clear();
  energy_obj.clear();

  // Release simulation states.  We don't care about the order.
  vector<Oxs_SimState*>::iterator sim_it = simulation_state.begin();
  while(sim_it!=simulation_state.end()) {
    delete (*sim_it);
    ++sim_it;
  }
  simulation_state.clear();
  simulation_state_reserve_count=0;
}

Oxs_Director::~Oxs_Director()
{
  Release();
  Tcl_DeleteExitHandler((Tcl_ExitProc *) ExitProc, (ClientData) this);
}

////////////////////////////////////////////////////////////////////////
// Tcl command creation/deletion interface.  Intended for use by
// Oxs_Ext objects.  The return value from RegisterTclCommand is NULL
// on error.  If non-NULL, the result be saved, to be eventually
// passed to DeleteTclCommand to remove the command from the
// interpreter.  DeleteTclCommand returns 0 on success, -1 on error.
// See also the documentation for the Tcl routines Tcl_CreateCommand,
// Tcl_DeleteCommand, and Tcl_GetCommandName.
Tcl_Command
Oxs_Director::RegisterTclCommand(const char* cmdname,
				 Oxs_Ext* instance) const
{
  // Should we add a check that we are not over-writing an
  // existing command?
  return Tcl_CreateCommand
    (interp,Oc_AutoBuf(cmdname),
     Oxs_ExtTclInterfaceProc,
     static_cast<ClientData>(instance),
     NULL);
}

int
Oxs_Director::DeleteTclCommand(Tcl_Command token) const
{
  return Tcl_DeleteCommand(interp,Tcl_GetCommandName(interp,token));
}


Oxs_DepKey<Oxs_Ext>
Oxs_Director::FindExtObject(const string& id) const
{
  // id should contain a colon (:).  If not, add a
  // trailing colon as a courtesy and match against
  // that.  (I.e., match against the class name with
  // the default instance name of "".)
  const string* idptr = &id;
  string scratch;
  if( id.find(':') == string::npos ) {
    scratch = id + ":";
    idptr = &scratch;
  }

  // Note: We use ext_map.find(id) rather than
  // ext_map[id] because the latter will insert a
  // dummy element into the map if 'id' is not in
  // the map.
  map<string,Oxs_Ext*>::const_iterator p = ext_map.find(*idptr);
  if( p == ext_map.end() )
    return Oxs_DepKey<Oxs_Ext>();
  return Oxs_DepKey<Oxs_Ext>(p->second);
}


string Oxs_Director::ListExtObjects() const
{
  string result;
  vector<Oxs_Ext*>::const_iterator it = ext_obj.begin();
  if(it!=ext_obj.end()) 
    result.append((*it)->InstanceName());
  while((++it)!=ext_obj.end()) {
    result.append("\n");
    result.append((*it)->InstanceName());
  }
  return result;
}

string Oxs_Director::ListEnergyObjects() const
{
  string result;
  vector<Oxs_Energy*>::const_iterator it = energy_obj.begin();
  if(it!=energy_obj.end()) 
    result.append((*it)->InstanceName());
  while((++it)!=energy_obj.end()) {
    result.append("\n");
    result.append((*it)->InstanceName());
  }
  return result;
}


// GetTotalEnergyAndmxH: Note that the return is mxH, as opposed to MxH.
// This relieves this routine from needing to know what Ms is, and saves
// an unneeded multiplication (since the evolver is just going to divide
// it back out again to calculate dm/dt (as opposed again to dM/dt)).
// The returned energy array is average energy density for the
// corresponding cell in J/m^3; mxH is in A/m.
void Oxs_Director::GetTotalEnergyAndmxH
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& mxH
 ) const
{
  // Allocate scratch space as needed
  temp_energy.AdjustSize(state.mesh);
  temp_field.AdjustSize(state.mesh);

  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  /// For convenience

  UINT4m i;
  UINT4m vecsize = state.mesh->Size();
  vector<Oxs_Energy*>::const_iterator it = energy_obj.begin();
  if(it!=energy_obj.end()) {
    (*it)->GetEnergyAndField(state,temp_energy,temp_field);
    for(i=0;i<vecsize;++i) energy[i] = temp_energy[i];
    for(i=0;i<vecsize;++i) {
      temp_field[i] ^= spin[i];
      mxH[i] = -1 * temp_field[i];   // mxH = -Hxm
    }
    while((++it)!=energy_obj.end()) {
      (*it)->GetEnergyAndField(state,temp_energy,temp_field);
      for(i=0;i<vecsize;++i) energy[i] += temp_energy[i];
      for(i=0;i<vecsize;++i) {
	temp_field[i] ^= spin[i];
	mxH[i] -= temp_field[i];     // mxH = -Hxm
      }
    }
  } else {
    // No energy terms!
    for(i=0;i<vecsize;++i) {
      energy[i] = 0.0;
      mxH[i].Set(0.,0.,0.);
    }
  }
}

// SIMULATION STATE MANAGEMENT: During construction, each Oxs_Ext object
// that wants to create and/or hold simulation states should register
// its intent with the director using ReserveSimulationStateRequest().
// The states are actually constructed only as needed, but the count
// obtained via reserve calls is used as an upper bound to protect
// against memory leaks swallowing up all of system memory.  Objects
// obtain keys (handles) to new states with the GetNewSimulationState()
// call.  The key initially holds a write lock on the object.  The
// caller can change the lock state, make copies, etc., as desired.  The
// GetNewSimulationState() routine scans through the list of states held
// in the simulation_state array, looking for a pre-created state that
// has no locks on it.  If one is found, a key is returned for that
// state.  If all existing states in the simulation_state array are
// locked, then a new state is created (subject to the max state count
// obtained from ReserveSimulationStateRequest() calls).  In all cases
// the returned key holds a write lock on the state.  The states are
// actually deleted only inside the director Release() command, after
// destruction of all Oxs_Ext objects.
//   NOTE: The key passed in to GetNewSimulationState() is allowed to
// hold a lock on an Oxs_SimState, *IF* that state was obtained from a
// previous call to GetNewSimulationState().  On entry,
// GetNewSimulationState() releases any lock currently held by the
// import key, but then immediately tests if the corresponding
// Oxs_SimState is then free.  If so, then GetNewSimulationState()
// resets the key to the same object.  Therefore, if the chance is good
// that the input key holds the sole lock on an object, then the lock
// should *not* be released before calling GetNewSimulationState().

void Oxs_Director::ReserveSimulationStateRequest
(UINT4m reserve_count)
{
  simulation_state_reserve_count+=reserve_count;
}

void Oxs_Director::GetNewSimulationState
(Oxs_Key<Oxs_SimState>& write_key)
{
  // Check first to see if import write_key is currently holding a lock.
  // If so, release it, and immediately check to see if that SimState
  // is available.
  const Oxs_SimState* sptr = write_key.GetPtr();
  if(sptr!=NULL) {
    write_key.Release();
    if( ! sptr->IsLocked() ) {
      write_key.Set(const_cast<Oxs_SimState*>(sptr));
      write_key.GetWriteReference();
      return;
    }
  }

  // Otherwise, go to the simulation_state list, and try to find
  // an unlocked state, working from the back forward.
  vector<Oxs_SimState*>::reverse_iterator it
    = simulation_state.rbegin();
  while(it!=simulation_state.rend()) {
    if(! (*it)->IsLocked()) {
      write_key.Set(*it);
      write_key.GetWriteReference();
      return;
    }
    ++it;
  }

  // If we get to here, then all existing states have locks
  // on them.  So create a new state.
  if(simulation_state.size()>=simulation_state_reserve_count) {
    OC_THROW(Oxs_BadResourceAlloc,
	     "Simulation state reserve count exhausted.");
  }
  Oxs_SimState* newstate = new Oxs_SimState;
  simulation_state.push_back(newstate);
  write_key.Set(newstate);
  write_key.GetWriteReference();
}
