/* FILE: director.h                 -*-Mode: c++-*-
 *
 * 3D solver director class.
 *
 */

#ifndef _OXS_DIRECTOR
#define _OXS_DIRECTOR

#include <vector>
#include <map>
#include <string>

#include "oc.h"
#include "depkey.h"
#include "key.h"
#include "simstate.h"
#include "mesh.h"
#include "meshvalue.h"
#include "threevector.h"

OC_USING_NAMESPACE_STD;  // Specify std namespace, if supported

/* End includes */

class Oxs_Ext;       // Forward declarations
class Oxs_Energy;

// 3D solver director
class Oxs_Director {
private:
  static void ExitProc(ClientData);
  int problem_loaded; // Used to prevent access to ProbRun
  /// until a valid problem is loaded.

  Tcl_Interp* const interp;

  void Release();
  vector<Oxs_Ext*> ext_obj;  // Ordered list of instantiated Oxs_Ext
  /// objects.  This list "owns" the objects, and is responsible for
  /// their destruction.
  map<string,Oxs_Ext*> ext_map; // Mapping of Ext names to objects
  vector<Oxs_Energy*> energy_obj; // Ordered list of energy terms.  This
  // is a sublist of ext_obj.

  // Array of held simulation states.  See discussion below concerning
  // ReserveSimulationStateRequest() and GetNewSimulationState().  This
  // array "owns" the simulations states, i.e., is responsible for their
  // destruction.
  UINT4m simulation_state_reserve_count;
  vector<Oxs_SimState*> simulation_state;

  // NB: Because ext_map and energy_obj hold duplicate pointers to
  // objects stored in ext_obj, one has to be careful to release
  // all pointers at the same time.  This is handled inside the 
  // Release() function.  We could store Oxs_DepKeys in ext_map
  // and energy_obj instead of pointers; this would provide notification
  // of hanging pointers, but since Oxs_Director is the owner of
  // these objects, at present I don't think this is necessary. -mjd

  mutable Oxs_MeshValue<REAL8m> temp_energy;
  mutable Oxs_MeshValue<ThreeVector> temp_field;

public:
  Oxs_Director(Tcl_Interp* i);
  ~Oxs_Director();  // Non-virtual.  Oxs_Director is not intended to
  // have children.

  int VerifyInterp(Tcl_Interp* i) { return (i==interp ? 1 : 0); }

  int ProbInit(const char* mif);
  int ProbReset();
  int ProbRelease() { Release(); return 1; } // Mainly for debugging

  // Tcl command creation/deletion interface.  Intended for use by
  // Oxs_Ext objects.  The return value from RegisterTclCommand is NULL
  // on error.  If non-NULL, the result be saved, to be eventually
  // passed to DeleteTclCommand to remove the command from the
  // interpreter.  DeleteTclCommand returns 0 on success, -1 on error.
  // See also the documentation for the Tcl routines Tcl_CreateCommand
  // and Tcl_DeleteCommandFromToken.
  Tcl_Command RegisterTclCommand(const char* cmdname,
				 Oxs_Ext* instance) const;
  int DeleteTclCommand(Tcl_Command token) const;

  Oxs_DepKey<Oxs_Ext> FindExtObject(const string& id) const;
  /// Returns a key pointing to the element of ext_map with instance id
  /// 'id'.  Returns an invalid key if the requested object is not
  /// found.  All keys handed out must release all locks before the
  /// corresponding object in ext_map can be deleted.  This means that
  /// any long-term locks held by ext_map objects to other ext_map
  /// objects should be only to objects earlier in the ext_obj list
  /// than themselves.  Perhaps there should be some type of access
  /// control on this function?

  string ListExtObjects() const;  // Mainly for debugging.
  string ListEnergyObjects() const;  // Mainly for debugging.

  void GetTotalEnergyAndmxH(const Oxs_SimState& state,
			    Oxs_MeshValue<REAL8m>& energy,
			    Oxs_MeshValue<ThreeVector>& mxH
			    ) const;
  /// Note that the return is mxH, as opposed to MxH.  This
  /// relieves this routine from needing to know what Ms is,
  /// and saves an unneeded multiplication (since the evolver
  /// is just going to divide it back out again to calculate
  /// dm/dt (as opposed again to dM/dt)).  The returned energy
  /// array is average energy density for the corresponding cell
  /// in J/m^3; mxH is in A/m.

  // SIMULATION STATE MANAGEMENT: During construction, each Oxs_Ext
  // object that wants to create and/or hold simulation states should
  // register its intent with the director using
  // ReserveSimulationStateRequest().  The states are actually
  // constructed only as needed, but the count obtained via reserve
  // calls is used as an upper bound to protect against memory leaks
  // swallowing up all of system memory.  Objects obtain keys (handles)
  // to new states with the GetNewSimulationState() call.  The key
  // initially holds a write lock on the object.  The caller can change
  // the lock state, make copies, etc., as desired.  The
  // GetNewSimulationState() routine scans through the list of states
  // held in the simulation_state array, looking for a pre-created state
  // that has no locks on it.  If one is found, a key is returned for
  // that state.  If all existing states in the simulation_state array
  // are locked, then a new state is created (subject to the max state
  // count obtained from ReserveSimulationStateRequest() calls).  In all
  // cases the returned key holds a write lock on the state.  The states
  // are actually deleted only inside the director Release() command,
  // after destruction of all Oxs_Ext objects.
  //   NOTE: The key passed in to GetNewSimulationState() is allowed to
  // hold a lock on an Oxs_SimState, *IF* that state was obtained from a
  // previous call to GetNewSimulationState().  On entry,
  // GetNewSimulationState() releases any lock currently held by the
  // import key, but then immediately tests if the corresponding
  // Oxs_SimState is then free.  If so, then GetNewSimulationState()
  // resets the key to the same object.  Therefore, if the chance is
  // good that the input key holds the sole lock on an object, then the
  // lock should *not* be released before calling
  // GetNewSimulationState().
  void ReserveSimulationStateRequest(UINT4m reserve_count);
  void GetNewSimulationState(Oxs_Key<Oxs_SimState>& write_key);

};

#endif // _OXS_DIRECTOR
