/* FILE: driver.cc                 -*-Mode: c++-*-
 *
 * Abstract driver class; fills states and steps evolvers.
 * Also registers the command "Oxs_Run" with the Tcl intepreter.
 *
 */

#include "oc.h"
#include "nb.h"
#include "director.h"
#include "driver.h"
#include "simstate.h"
#include "evolver.h"
#include "key.h"
#include "energy.h"	// Needed to make MSVC++ 5 happy

/* End includes */

// Constructor
Oxs_Driver::Oxs_Driver
( const char* name,        // Child instance id
  Oxs_Director* newdtr,    // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr       // Args; should include an 
                          /// "evolver <evolver-name>" pair.
  ) : Oxs_Ext(name,newdtr,safe_interp,argstr)
{
  // Get evolver name specification
  CheckInitValueParamCount("evolver",1);
  const string& evolver_name = (*FindInitValue("evolver"))[0];

  // Put dependency lock on specified evolver
  Oxs_DepKey<Oxs_Ext> key = director->FindExtObject(evolver_name);
  evolver.Set(dynamic_cast<Oxs_Evolver*>(key.GetPtr()));
  if(evolver.GetPtr()==NULL) {
    string msg="No Oxs_Evolver object found with name "
      + evolver_name;
    throw Oxs_Ext::Error(msg.c_str());
  }
  DeleteInitValue("evolver");
  // evolver holds a dependency lock on the Oxs_Evolver
  // that won't be freed until *this is destroyed.

  // Reserve state space in director
  director->ReserveSimulationStateRequest(2);

  // Register Oxs_Run command
  tclcmd_token=director->RegisterTclCommand("Oxs_Run",this);
}

//Destructor
Oxs_Driver::~Oxs_Driver()
{
  director->DeleteTclCommand(tclcmd_token);
  tclcmd_token=NULL;
}

// Oxs_Driver version of Init().  All children of Oxs_Driver *must*
// call this function in their Init() routines.  The main purpose
// of this function is to initialize the current state.
BOOL Oxs_Driver::Init_Driver()
{
  // Initialize current state from initial state provided by
  // concrete child class.
  current_state = GetInitialState();
  return (current_state.GetPtr() != NULL);
}

int Oxs_Driver::TclInterface
(Tcl_Interp* interp, int argc, char** argv)
{
  Tcl_ResetResult(interp);

  const char* meshtype = "rectangular";  // Either rectangular or irregular
  /// This should really be set at runtime.

  if(current_state.GetPtr() == NULL) {
    // Current state is not initialized.
    string msg="Current state in Oxs_Driver is not initialized;"
      " This is probably the fault of the child class "
      + string(ClassName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  if (argc > 1) {
    char c = argv[1][0];
    if (c == 'D') {
      if (strcmp(argv[1], "DumpMagnetization") == 0) {
	if (argc != 4) {
	  Tcl_AppendResult(interp, "missing args: temp final", (char *) NULL);
	  return TCL_ERROR;
	}
	
      // Dump final spin state to temp file
      char buf[1024];
      const Oxs_SimState& sref = current_state.GetReadReference();
      Oc_Snprintf(buf,sizeof(buf),"%s-%d.omf",argv[3],sref.iteration_count);
      sref.mesh->WriteOvf
        (argv[2],buf,
         "This is debug output from the prototype oxs solver",
         meshtype,"binary","8",
         &sref.spin,NULL);
      Tcl_AppendElement(interp,buf);
      return TCL_OK;

      }
    }
    if (c == 'M') {
      if (strcmp(argv[1], "MakeDataTable") == 0) {
	if (argc != 2) {
	  Tcl_AppendResult(interp,
			   "MakeDataTable subcommand takes no args",
			   (char *) NULL);
	  return TCL_ERROR;
	}
	
      // Create datatable-style string of triples
      char buf[1024];
      const Oxs_SimState& sref = current_state.GetReadReference();
      Oc_Snprintf(buf,sizeof(buf),"%s %s %d",
		  "Stage","{}",sref.stage_number);
      Tcl_AppendElement(interp,buf);
      Oc_Snprintf(buf,sizeof(buf),"%s %s %d",
		  "Iteration","{}",sref.iteration_count);
      Tcl_AppendElement(interp,buf);
      Oc_Snprintf(buf,sizeof(buf),"%s %s %.16g",
		  "{Step size}","ns",sref.last_timestep*1e9);
      Tcl_AppendElement(interp,buf);
      Oc_Snprintf(buf,sizeof(buf),"%s %s %.16g",
		  "{Sim Time}","ns",sref.elapsed_time*1e9);
      Tcl_AppendElement(interp,buf);
      Oc_Snprintf(buf,sizeof(buf),"%s %s %.16g",
		  "{Total Energy}","J",sref.total_energy);
      Tcl_AppendElement(interp,buf);
      Oc_Snprintf(buf,sizeof(buf),"%s %s %.16g",
		  "dm/dt","deg/ns",sref.max_dm_dt*180.0*1e-9/PI);
      Tcl_AppendElement(interp,buf);

      // Reduced magnetization
      ThreeVector m = sref.mesh->VolumeSum(&sref.spin); // Each spin
      /// should be weighted by corresponding Ms, but VolumeSum takes
      /// an Oxs_MeshValue<REAL8m> as the optional second argument, and
      /// all that is currently available is the Oxs_FixedScalarField
      /// pointer stored in current_state.  Oxs_SimState should probably
      /// migrate to using Oxs_MeshValue instead of Oxs_FixedScalarField.
      REAL8m total_volume = sref.mesh->TotalVolume();
      if(total_volume>0.) m *= (1./total_volume);
      Oc_Snprintf(buf,sizeof(buf),"%s %s %.16g",
		  "Mx/Ms","{}",m.x);
      Tcl_AppendElement(interp,buf);
      Oc_Snprintf(buf,sizeof(buf),"%s %s %.16g",
		  "My/Ms","{}",m.y);
      Tcl_AppendElement(interp,buf);
      Oc_Snprintf(buf,sizeof(buf),"%s %s %.16g",
		  "Mz/Ms","{}",m.z);
      Tcl_AppendElement(interp,buf);

      return TCL_OK;
      }
    }
  }

  if (Done(current_state.GetReadReference(), interp, argc, argv)) {
    const Oxs_SimState& sref = current_state.GetReadReference();
    sref.mesh->WriteOvf
      ("oxs_output.omf","Test oxs output",
       "This is debug output from the prototype oxs solver",
       meshtype,"binary","8",
       &sref.spin,NULL);
    return TCL_OK;
  }

  // Repeat forever, until Done() or Return() breaks us out
  while (1) {

    // Attempt to take a step.
    Oxs_Key<Oxs_SimState> next_state;
    director->GetNewSimulationState(next_state);
    FillStatePartA(current_state.GetReadReference(),
                   next_state.GetWriteReference());
    if( (*evolver.GetPtr()).Step(*this,current_state,next_state) ) {
      // Good step.  Release read lock on old current_state,
      // and copy key from next_state.
      current_state = next_state;
  
      // Give interface a chance to know we have a new state.
      NewState(current_state.GetReadReference(), interp, argc, argv);
    }
  
    if (StageDone(current_state.GetReadReference(), interp, argc, argv)) {
      director->GetNewSimulationState(next_state);
      Oxs_SimState& nstate = next_state.GetWriteReference();
      nstate = current_state.GetReadReference();
      nstate.valid_state=0; // max_dm_dt is probably incorrect.
      // For this reason keep the same iteration_count.
      nstate.stage_iteration_count=0;
      nstate.stage_number++;
      next_state.GetReadReference();  // Release write lock
      current_state = next_state;
    }

    if (Done(current_state.GetReadReference(), interp, argc, argv)) {
      char buf[1024];
      const Oxs_SimState& sref = current_state.GetReadReference();
      const ThreeVector v = sref.spin[0];
      Oc_Snprintf(buf,sizeof(buf),
	      "Max dm_dt: %g, spin=(%g,%g,%g)\n"
	      "Total iteration count: %d\n"
	      "Stage iteration count: %d\n"
	      "Stage number: %d\n"
	      "Total Energy: %g",
	      sref.max_dm_dt,v.x,v.y,v.z,
	      sref.iteration_count,
	      sref.stage_iteration_count,
	      sref.stage_number,
	      sref.total_energy);
      Tcl_AppendElement(interp,buf);
      return TCL_OK;
    }

    if (Return(current_state.GetReadReference(), interp, argc, argv)) {
      return TCL_OK;
    }
  }
}

