/* FILE: ext.cc                 -*-Mode: c++-*-
 *
 * Base OXS extension class.  This is an abstract class.
 *
 */


#include <iostream>
#include "oc.h"
#include "ext.h"
#include "util.h"

/* End includes */

Oxs_Ext::~Oxs_Ext() {}

Oxs_Ext::Oxs_Ext(const char* idname,Oxs_Director* dtr)
  : instanceId(idname), director(dtr), tclcmd_token(NULL) {}

// Extended constructor with built-in argument parsing
Oxs_Ext::Oxs_Ext(const char* idname,Oxs_Director* dtr,
                 Tcl_Interp* /* safe_interp */,const char* argstr)
  : instanceId(idname), director(dtr), tclcmd_token(NULL)
{
  // Use Tcl_SplitList to break argstr into components
  Oxs_SplitList arglist;
  if(arglist.Split(argstr)!=TCL_OK) {
    // Input string not a valid Tcl list; Throw exception
    string msg="MIF input block for "
      + string(InstanceName())
      + " not a valid Tcl list: "
      + string(argstr);
    throw Oxs_Ext::Error(msg.c_str());
  }
  if(arglist.Count()%2 != 0) {
    // Input string is not a list of pairs; Throw exception
    string msg="MIF input block for "
      + string(InstanceName())
      + " does not have an even number of elements: "
      + string(argstr);
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Fill in init_strings key-value array
  Oxs_SplitList paramlist;
  for(int i=0;i+1<arglist.Count();i+=2) {
    // Split up parameter list
    if(paramlist.Split(arglist[i+1])!=TCL_OK) {
      // Input string not a valid Tcl list; Throw exception
      string msg="MIF input block for "
	+ string(InstanceName())
	+ " has item that is not a valid Tcl list: "
	+ string(arglist[i+1]);
      throw Oxs_Ext::Error(msg.c_str());
    }

    // Store key + list pair
    const string& key = arglist[i];
    init_strings[key].clear();
    for(int j=0;j<paramlist.Count();j++) {
      init_strings[key].push_back(paramlist[j]);
    }
  }
}

const vector<string>*
Oxs_Ext::FindInitValue(const string& key)
{ // Returns parameter list corresponding to key, or a NULL pointer
  // if key is not in init_strings.  NOTE: The return value from this
  // routine will become invalid if DeleteInitValue is called with the
  // same key.
  map<string,vector<string> >::const_iterator it
    = init_strings.find(key);
  if(it==init_strings.end()) return NULL;
  return &(it->second);
}


void
Oxs_Ext::CheckInitValueParamCount(const string& key,UINT4m count)
{ // Throws an exception if key is not in init_strings, or if the number
  // of parameters in the corresponding value is not exactly count.
  const vector<string>* param = FindInitValue(key);
  if(param==NULL) {
    string msg = "MIF input block for "
      + string(InstanceName())
      + " missing specification for parameter \""
      + key + "\"";
      throw Oxs_Ext::Error(msg.c_str());
  }
  if(param->size()!=count) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
		"MIF input block %s, \"%s\" specification"
                " has %d items; it should have exactly %d items.",
		InstanceName(),key.c_str(),param->size(),count);
    throw Oxs_Ext::Error(buf);
  }
}

BOOL
Oxs_Ext::DeleteInitValue(const string& key)
{ // Removes the specified key-value pair from init_strings.
  //  Returns 1 on success, 0 if key was not found in init_strings.
  return static_cast<BOOL>(init_strings.erase(key));
}

void Oxs_Ext::VerifyAllInitArgsUsed()
{ // Throws an exception if GetInitValueCount() != 0.  Use this in
  // combination with DeleteInitValue() to check MIF input block for
  // superfluous parameters.  This should be called only the actual
  // instantiating class (not a parenting class).
  if(GetInitValueCount()!=0) {
    string msg = "MIF input block for \""
      + string(InstanceName())
      + "\" includes the following unrecognized parameters:\n";
    map<string,vector<string> >::const_iterator it
      = init_strings.begin();
    while(it!=init_strings.end()) {
      msg += "   " + it->first;
      for(UINT4m i=0; i < it->second.size(); i++) {
	msg += " " + it->second[i];
      }
      msg += "\n";
      ++it;
    }
    throw Oxs_Ext::Error(msg.c_str());
  }
}

string Oxs_Ext::ExtractClassName(const string& name)
{
  string classname=name;
  string::size_type endpt=classname.rfind(':');
  if(endpt!=string::npos) {
    classname.resize(endpt);
  } else {
    classname=""; // No colon found
  }
  return classname;
}

BOOL Oxs_Ext::ValidateIdStrings() const
{
  string classrequest=ExtractClassName(instanceId);
  if(classrequest.compare(ClassName())!=0) {
    // Error, wrong type!
    string msg="Incorrect object type request ("
      + classrequest
      + ") in "
      + ClassName()
      + " constructor.";
    throw Oxs_Ext::Error(msg.c_str());
  }
  return 1;
}


// Oxs_Ext Registration code /////////////////////////////////////

Oxs_ExtRegistrationBlock::Oxs_ExtRegistrationBlock
(const char* name,Oxs_ExtChildConstructFunc* ctr)
  : classid(name), construct(ctr)
{
  Oxs_Ext::Register(*this);
}

// Function that returns a reference to an STL "map" of all registered
// classes.  The map is a static variable of RegClasses(). This insures
// the map is initialized before it is used, which would otherwise be
// difficult, because Oxs_Ext child classes call Oxs_Ext::Register
// during static initialization.  (See Section 9.4.1 of "The C++
// Programming Lanuage," Bjarne Stroustrup, 3rd edition for a discussion
// of nonlocal variable initialization order.)
map<string,Oxs_ExtRegistrationBlock>& Oxs_Ext::RegClasses()
{
  static map<string,Oxs_ExtRegistrationBlock> regclasses;
  return regclasses;
}

void Oxs_Ext::Register(Oxs_ExtRegistrationBlock const& regblk)
{
  map<string,Oxs_ExtRegistrationBlock>& regclasses=RegClasses();
  regclasses[regblk.classid]=regblk;
}

Oxs_Ext* Oxs_Ext::MakeNew
(const char* idstring, // Child object name
 Oxs_Director* dtr,    // App director
 Tcl_Interp* safe_interp, // Safe interpreter
 const char* argstr)   // MIF input block params
{
  map<string,Oxs_ExtRegistrationBlock>& regclasses=RegClasses();
  string classname(ExtractClassName(idstring));
  map<string,Oxs_ExtRegistrationBlock>::iterator it =
    regclasses.find(classname);
  if(it==regclasses.end()) return NULL;
  Oxs_Ext* obj=NULL;
  try {
    obj=it->second.construct(idstring,dtr,safe_interp,argstr);
    obj->ValidateIdStrings();  // Verify we got the right one! (Safety)
  }
  catch (Oxs_Ext::Error err) {
    cerr << "Oxs_Ext construction error: " << err.msg << "\n" ;
    throw;
  }
  return obj;
}

string Oxs_Ext::ListRegisteredChildClasses()
{
  map<string,Oxs_ExtRegistrationBlock>& regclasses=RegClasses();
  string result("Oxs Registration List ---");
  map<string,Oxs_ExtRegistrationBlock>::iterator it;
  for(it=regclasses.begin();it!=regclasses.end();++it) {
    result.append("\n ");
    result.append(it->first);
  }
  return result;
}

// Tcl interface support code //////////////////////////////////////////
extern "C" int
Oxs_ExtTclInterfaceProc(ClientData cd,Tcl_Interp *interp,
			int argc,char **argv)
{
  int result=TCL_ERROR;
  try {
    Oxs_Ext* instance=static_cast<Oxs_Ext*>(cd);
    result = instance->TclInterface(interp,argc,argv);
  }
  catch (Oxs_Ext::Error err) {
    cerr << "Oxs_Ext runtime error: " << err.msg << "\n" ;
    throw;
  }
  return result;
}

int Oxs_Ext::TclInterface(Tcl_Interp*,int,char**)
{ // Default implementation is a NOP
  return TCL_OK;
}


////////////////////////////////////////////////////////////////////////
