/* FILE: ext.h                 -*-Mode: c++-*-
 *
 * Base OXS extension class.  This is an abstract class.
 *
 */

#ifndef _OXS_EXT
#define _OXS_EXT

#include <map>
#include <string>
#include <vector>

#include "oc.h"
#include "depkey.h"
#include "deplock.h"

OC_USING_NAMESPACE_STD;  // Specify std namespace, if supported

/* End includes */

// Macro that builds a generic child constructor, and provides a static
// Oxs_ExtRegistrationBlock variable to force automatic class
// registration.  This macro should be invoked at the top of each
// Oxs_Ext child class *.cc class definition file that requires
// registration, i.e., any class for which object instantiation is
// desired.
//   The child constructor produced by this macro has the name
//
//          OxsExtNewChild##classname
//
// where "classname" is the macro argument, and the name of the Oxs_Ext
// child class being registered.
//   This macro also provides definitions for the class public member
// function ClassName().  This is used instead of an RTTI call, because
// we want to be able to associate this name reported by the class with
// entries in the input MIF file.  The RTTI "typeid().name()" function
// returns an implementation dependent string, which is not what we want.
#define OXS_EXT_REGISTER(classname)                               \
static Oxs_Ext*                                                   \
OxsExtNewChild##classname(const char* instancename,               \
			  Oxs_Director* newdtr,                   \
                          Tcl_Interp* safe_interp,                \
			  const char* argstr)                     \
{ return static_cast<Oxs_Ext*>(                                   \
     new classname(instancename,newdtr,safe_interp,argstr)); }    \
const char* classname::ClassName() const { return #classname; }   \
static Oxs_ExtRegistrationBlock oxsExt##classname##RegBlock(      \
#classname,&OxsExtNewChild##classname)

class Oxs_Ext;       // Forward declarations
class Oxs_Director;
extern "C" Tcl_CmdProc Oxs_ExtTclInterfaceProc;

typedef Oxs_Ext* Oxs_ExtChildConstructFunc(
   const char* instancename, // Child instance id string
   Oxs_Director* dtr,        // App director
   Tcl_Interp* safe_interp,  // Safe interpreter
   const char* argstr);      // MIF input block parameters

// Child class registration block
struct Oxs_ExtRegistrationBlock {
public:
  string classid;  // Class name
  Oxs_ExtChildConstructFunc * construct;  // Constructor

  // Standard constructor.  Performs automatic registration
  // by calling Oxs_Ext::Register();
  Oxs_ExtRegistrationBlock(const char* name,
			   Oxs_ExtChildConstructFunc* ctr);

  // Default constructor, for use by STL "map" container.
  Oxs_ExtRegistrationBlock():classid(""),construct(NULL) {}
};


////////////////////////////////////////////////////////////////////////
// Oxs_Ext base class
class Oxs_Ext : public Oxs_DepLock {
  // Perhaps the inheritance from Oxs_DepLock should be private,
  // but then exposing the Oxs_DepLock member functions is awkward.
  friend struct Oxs_ExtRegistrationBlock; // Makes ::Register() calls.
private:
  // List (actually, "map")  of all registered classes.  The actual
  // map is stored as a function static variable of RegClasses().
  // This insures the map is initialized before it is used, which
  // is difficult to insure otherwise, because Oxs_Ext child classes
  // call Oxs_Ext::Register during static initialization.  (See
  // Section 9.4.1 of "The C++ Programming Lanuage," Bjarne Stroustrup,
  // 3rd edition for a discussion of nonlocal variable initialization
  // order.)
  static map<string,Oxs_ExtRegistrationBlock>& RegClasses();

  static void Register(const Oxs_ExtRegistrationBlock & regblk);

  const string instanceId;     // Object instance name

  // Disable copy constructor and assignment operators.  Don't
  // provide definitions for these!
  Oxs_Ext(const Oxs_Ext&);
  Oxs_Ext& operator=(const Oxs_Ext&);

protected:
  Oxs_Director* director;  // Controlling director object

  // Base Oxs_Ext constructor.  Call this from a child class if
  // the child wants to parse the MIF argstr on its own. The
  // extended Oxs_Ext constructor farther below is available to
  // provide automatic MIF argstr parsing.
  Oxs_Ext(const char* idname,Oxs_Director* dtr);

  // Helper routine to extract className from an instanceName
  static string ExtractClassName(const string& name);

  // Routine to check that instanceId and classId are consistent.
  // Throws an error if they aren't.  This is called inside
  // Oxs_Ext::MakeNew().
  BOOL ValidateIdStrings() const;

  // Tcl interface support code
  friend Tcl_CmdProc Oxs_ExtTclInterfaceProc;
  Tcl_Command tclcmd_token;
  virtual int TclInterface(Tcl_Interp* interp,int argc,char** argv);
  /// Define this if you want to register a command in a Tcl
  /// interpreter.  Use tclcmd_token to remove the command from
  /// the interpreter, in general but specifically on instance
  /// destruction.

  // Support code for child construction.  This code is activated
  // by the use of the following constructor (which has an extended
  // parameter list including 'const char* argstr').  This constructor
  // requires that argstring be a Tcl list of pairs, where the first
  // element in each pair is an id string, and the second is a Tcl
  // list of parameters.  The id string is mapped to the key entry
  // in init_strings, and the parameter list is broken down and stored
  // as the value corresponding to that key.  If the child MIF class
  // argstr does not fit this format, then use the base Oxs_Ext
  // constructor above.
  Oxs_Ext(const char* idname,Oxs_Director* dtr,Tcl_Interp* safe_interp,
	  const char* argstr);

  map<string,vector<string> > init_strings; // Initialization strings.

  const vector<string>* FindInitValue(const string& key); // Returns
  /// parameter list corresponding to key, or a NULL pointer if key is
  /// not in init_strings.  NOTE: The return value from this routine
  /// will become invalid if DeleteInitValue is called with the same
  /// key.

  void CheckInitValueParamCount(const string& key,UINT4m count);
  /// Throws an exception if key is not in init_strings, or if
  /// the number of parameters in the corresponding value is not
  /// exactly count.

  BOOL DeleteInitValue(const string& key); // Removes the specified
  /// key-value pair from init_strings.  Returns 1 on success, 0
  /// if key was not found in init_strings.

  UINT4m GetInitValueCount() {
    return static_cast<UINT4m>(init_strings.size());
  }

  void VerifyAllInitArgsUsed(); // Throws an exception if
  /// GetInitValueCount() != 0.  Use this in combination with
  /// DeleteInitValue() to check MIF input block for superfluous
  /// parameters.  This should be called only the actual
  /// instantiating class (not a parenting class).

public:
  struct Error {
    const char* msg;
    Error(const char* errmsg) : msg(errmsg) {}
  };
  const char* InstanceName() const { return instanceId.c_str(); }
  virtual const char* ClassName() const =0;    // Child class name

  virtual BOOL Init() =0;  // This can be called at any time to reset
  /// an Ext object to its initial state.  The constructor should set
  /// all static object state information, and then call this function
  /// to initialize any state information that might change during the
  /// lifetime of the object.  It is incumbent upon the constructor
  /// to save whatever information is necessary for Init() to completely
  /// reset itself.

  static Oxs_Ext* MakeNew(const char* idstring, // Child object name
			  Oxs_Director* dtr,    // App director
                          Tcl_Interp* safe_interp, // Safe interpreter
			  const char* argstr);  // MIF input block params
  /// NOTE: The safe interpreter is not guaranteed valid beyond the
  /// call to the constructor.  It is the interpreter used by the Oxs_Mif
  /// object to source the input MIF file.


  static string ListRegisteredChildClasses(); // This is intended
  /// for debugging purposes only.

  virtual ~Oxs_Ext();
};


#endif // _OXS_EXT
