/* FILE: key.h                 -*-Mode: c++-*-
 *
 * Oxs_Key class, intended for use with the Oxs_Lock family.
 *
 * Usage notes:  In general, in passing around pointers to Oxs_Lock
 * objects, one has 3 choices: pass a raw pointer, pass an
 * Oxs_ConstKey object, or pass an Oxs_Key object.  Here are some
 * general guidelines (T below represents the Oxs_Lock subclass):
 *
 *    1) const T* : Pass a 'const T* ' when the receiving routine needs
 *        immediate but not long-term read access to T.  The point here
 *        is that setting a read lock on T requires write access to T.
 *
 *    2) Oxs_ConstKey<T> :  Pass an 'Oxs_ConstKey<T>' object by value
 *        when the receiving routine needs only read access to T, but
 *        may want to cache T for future (i.e., after return) use as
 *        well.  Because Oxs_ConstKey<T> is not "const", the callee
 *        can gain a read lock, if desired.  The object should usually
 *        be passed by value (as opposed to reference) so the caller
 *        can invoke with a Oxs_Key<T> object.
 *
 *    3) Oxs_Key<T>& : Pass an 'Oxs_Key<T>' object by reference when the
 *        receiving routine needs to write to the object.  This *must*
 *        be a 'by reference' call, because to obtain a write lock there
 *        can be only 1 key pointing to the object.  The receiving
 *        routine should wrap the code with the outstanding write lock
 *        in a try block to be certain the write lock gets released.
 *
 *   4) Oxs_Key<T> : An 'Oxs_Key<T>' object may be passed by value as
 *        the return value from a function.  However, because this
 *        involves a copy constructor in the caller, the key must be
 *        set to the read state.  Some compilers may optimize away
 *        this copy, but one should not count on it.  Moreover, some
 *        compilers may be lazy about (i.e., procrastinate) destroying
 *        objects in the called function, so it is best not to put
 *        a write lock on a key returned by value.
 *
 * The key point about passing around an Oxs_Key<T> object is that the
 * receiver will be allowed by the compiler to try to place a write lock
 * on it.  If there is in existence another key pointing to the same
 * T object, then the lock attempt will fail...but this will only be
 * caught at runtime.  It is better to use the conventions outlined
 * above to more clearly specify the contract between the caller and
 * receiver, and to allow violations of that contract to be detected
 * during compilation.
 */

#ifndef _OXS_KEY
#define _OXS_KEY

#include "oc.h"
#include "oxsexcept.h"  // Needed in template definitions

/* End includes */

template<class T> class Oxs_ConstKey;  // Forward declaration

////////////////////////////////////////////////////////////////////////
// Key class.  Class T should be a child of Oxs_Lock.
// Compare also the Oxs_ConstKey class.
template<class T> class Oxs_Key
{
  friend class Oxs_ConstKey<T>;
private:
  UINT4m id;
  T* ptr;
  enum Lock_Held { READ, WRITE, INVALID } lock;
  // lock==INVALID iff id==0 and ptr==NULL
public:
  void Release(); // Release lock and go into the INVALID state.

  void Set(T* tptr); // Releases current locks, if any, sets ptr to tptr,
  /// and if tptr!=NULL tries to set a READ lock on it.  Throws an
  /// exception if READ lock attempt fails.

  Oxs_Key() : id(0), ptr(0), lock(INVALID) {}
  Oxs_Key(T* tptr) : id(0), ptr(0), lock(INVALID) { Set(tptr); }
  ~Oxs_Key() { Release(); }

  void Swap(Oxs_Key<T>& other);

  // Copy constructor.  Duplicates id and ptr info, and
  // creates a read lock.
  Oxs_Key(const Oxs_Key<T>& other);
  Oxs_Key(const Oxs_ConstKey<T>& other);

  Oxs_Key<T>& operator=(const Oxs_Key<T>& other);
  /// Same behavior as copy constructor, after releasing
  /// any locks on current object.

  UINT4m GetId() const { return id; }

  const T& GetReadReference();  // Release write lock if set,
  /// and sets a read lock.  Returns a read reference; if this
  /// key already holds a read lock, no other action occurs.

  T& GetWriteReference(); // Releases read lock if set, and sets
  /// a write lock.  Other than returning a write reference, this
  /// routine is a NOP if this key already holds a write lock.

  const T* GetPtr() const { return ptr; } // Returns pointer
  /// with no changes to lock.  A NULL indicates an unset state.
};


////////////////////////////////////////////////////////////////////////
// ConstKey class.  Class T should be a child of Oxs_Lock.
// This is just the Oxs_Key class above but without WriteLock access.
// Note that ConstKey holds a T* as opposed to a const T*.  This
// is because we need write access to T* to set even a read lock.
// Such a lock can be detected externally, so setting a read lock
// isn't really "conceptually constant" (See Item 21 of Scott Meyer's
// "Effective C++.")
// MAINTAINER NOTE: Be sure to mirror any changes between the two
// classes.
template<class T> class Oxs_ConstKey
{
  friend class Oxs_Key<T>;
private:
  UINT4m id;
  T* ptr;
  enum Lock_Held { READ, INVALID } lock;
  // lock==INVALID iff id==0 and ptr==NULL
public:
  void Release();
  void Set(T* tptr); // Releases current locks, if any, sets ptr to tptr,
  /// and if tptr!=NULL tries to set a READ lock on it.  Throws an
  /// exception if READ lock attempt fails.

  Oxs_ConstKey() : id(0), ptr(0), lock(INVALID) {}
  Oxs_ConstKey(T* tptr)
    : id(0), ptr(0), lock(INVALID) { Set(tptr); }
  ~Oxs_ConstKey() { Release(); }

  void Swap(Oxs_ConstKey<T>& other);

  // Copy constructor.  Duplicates id and ptr info, and
  // creates a read lock.
  Oxs_ConstKey(const Oxs_ConstKey<T>& other);
  Oxs_ConstKey(const Oxs_Key<T>& other);

  Oxs_ConstKey<T>& operator=(const Oxs_ConstKey<T>& other);
  /// Same behavior as copy constructor, after releasing
  /// any locks on current object.

  UINT4m GetId() const { return id; }

  const T& GetReadReference() const;  // Checks lock validity,
  // and returns *ptr.

  const T* GetPtr() const { return ptr; } // Returns pointer
  /// with no changes to lock.  A NULL indicates an unset state.
};


////////////////////////////////////////////////////////////////////////
// Include all definitions here.  This is required by compiler/linkers
// using what gcc calls the "Borland model," as opposed to the "Cfront
// model."  The latter is distinguished by the use of template
// repositories.  With the Cfront model, template definitions can be
// put in a file (*.cc) separate from the header file (*.h).  Putting
// the definitions in the header file along with declarations slows down
// compilation, because the definitions are re-compiled in each including
// file, but appears to be more portable.

// Oxs_Key class
template<class T>
void Oxs_Key<T>::Release()
{
  if(lock!=INVALID) {
    if(ptr==NULL)
      OC_THROW(Oxs_BadPointer,
	       "Oxs_Key<T>::Release(): NULL pointer");
    if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
    if(lock==READ) ptr->ReleaseReadLock();
    else if(lock==WRITE) ptr->ReleaseWriteLock();
    lock=INVALID;
  }
  id=0;
  ptr=NULL;
}

template<class T>
void Oxs_Key<T>::Set(T* tptr)
{
  Release();
  if(tptr!=NULL) {
    if(!tptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    ptr=tptr;
    id=ptr->Id();
    lock=READ;
  }
}

// Exchange lock objects
template<class T>
void Oxs_Key<T>::Swap(Oxs_Key<T>& other)
{ // Note: This works okay also if this==&other
  UINT4m      tmpid=id;      id=other.id;      other.id=tmpid;
  T*         tmpptr=ptr;    ptr=other.ptr;    other.ptr=tmpptr;
  Lock_Held tmplock=lock;  lock=other.lock;  other.lock=tmplock;
}

// Copy constructor.  Duplicates id and ptr info, and
// creates a read lock.  Will throw an exception if
// other is holding a write lock.
template<class T>
Oxs_Key<T>::Oxs_Key(const Oxs_Key<T>& other)
{
  if(other.lock==INVALID) {
    id=0;
    ptr=NULL;
    lock=INVALID;
  } else {
    id=other.id;
    ptr=other.ptr;
    if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
    if(!ptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    lock=READ;
  }
}

template<class T>
Oxs_Key<T>::Oxs_Key(const Oxs_ConstKey<T>& other)
{
  if(other.lock == Oxs_ConstKey<T>::INVALID) {
    id=0;
    ptr=NULL;
    lock=INVALID;
  } else {
    id=other.id;
    ptr=other.ptr;
    if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
    if(!ptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    lock=READ;
  }
}

template<class T>
Oxs_Key<T>& Oxs_Key<T>::operator=(const Oxs_Key<T>& other)
{ // Same behavior as copy constructor
  if(this==&other) return *this; // Nothing to do
  Release();
  if(other.lock!=INVALID) {
    id=other.id;
    ptr=other.ptr;
    if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
    if(!ptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    lock=READ;
  }
  return *this;
}

template<class T>
const T& Oxs_Key<T>::GetReadReference()
{ // Release write lock, if set, and sets a read lock.
  // A NOP if this key already holds a read lock.
  if(ptr==NULL)
    OC_THROW(Oxs_BadPointer,
	     "Oxs_Key<T>::GetReadReference(): NULL pointer");
  if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
  if(lock==WRITE) {
    ptr->ReleaseWriteLock();
    id=ptr->Id();
    if(!ptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    lock=READ;
  }
  return *ptr;
}

template<class T>
T& Oxs_Key<T>::GetWriteReference()
{ // Releases read lock and sets a write lock.
  // A NOP if this key already holds a write lock.
  if(ptr==NULL)
    OC_THROW(Oxs_BadPointer,
	     "Oxs_Key<T>::GetWriteReference(): NULL pointer");
  if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
  if(lock==READ) {
    ptr->ReleaseReadLock();
    if(!ptr->SetWriteLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire write lock");
    id=ptr->Id();
    lock=WRITE;
  }
  return *ptr;
}


// Oxs_ConstKey class
template<class T>
void Oxs_ConstKey<T>::Release()
{
  if(lock!=INVALID) {
    if(ptr==NULL)
      OC_THROW(Oxs_BadPointer,
	       "Oxs_ConstKey<T>::Release(): NULL pointer");
    if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
    ptr->ReleaseReadLock();
    lock=INVALID;
  }
  id=0;
  ptr=NULL;
}

template<class T>
void Oxs_ConstKey<T>::Set(T* tptr)
{
  Release();
  if(tptr!=NULL) {
    if(!tptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    ptr=tptr;
    id=ptr->Id();
    lock=READ;
  }
}

// Exchange lock objects
template<class T>
void Oxs_ConstKey<T>::Swap(Oxs_ConstKey<T>& other)
{ // Note: This works okay also if this==&other
  UINT4m      tmpid=id;      id=other.id;      other.id=tmpid;
  T*         tmpptr=ptr;    ptr=other.ptr;    other.ptr=tmpptr;
  Lock_Held tmplock=lock;  lock=other.lock;  other.lock=tmplock;
}


// Copy constructor.  Duplicates id and ptr info, and
// creates a read lock.
template<class T>
Oxs_ConstKey<T>::Oxs_ConstKey(const Oxs_ConstKey<T>& other)
{
  if(other.lock==INVALID) {
    id=0;
    ptr=NULL;
    lock=INVALID;
  } else {
    id=other.id;
    ptr=other.ptr;
    if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
    if(!ptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    lock=READ;
  }
}

template<class T>
Oxs_ConstKey<T>::Oxs_ConstKey(const Oxs_Key<T>& other)
{
  if(other.lock == Oxs_Key<T>::INVALID) {
    id=0;
    ptr=NULL;
    lock=INVALID;
  } else {
    id=other.id;
    ptr=other.ptr;
    if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
    if(!ptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    lock=READ;
  }
}

template<class T> Oxs_ConstKey<T>&
Oxs_ConstKey<T>::operator=(const Oxs_ConstKey<T>& other)
{ // Same behavior as copy constructor
  if(this==&other) return *this; // Nothing to do
  Release();
  if(other.lock!=INVALID) {
    id=other.id;
    ptr=other.ptr;
    if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
    if(!ptr->SetReadLock())
      OC_THROW(Oxs_BadLock,"Unable to acquire read lock");
    lock=READ;
  }
  return *this;
}

template<class T>
const T& Oxs_ConstKey<T>::GetReadReference() const
{ // Checks key validity, and returns a const reference.
  if(ptr==NULL)
    OC_THROW(Oxs_BadPointer,
	     "Oxs_ConstKey<T>::GetReadReference(): NULL pointer");
  if(id!=ptr->Id()) OC_THROW(Oxs_BadPointer,"Bad lock id");
  return *ptr;
}

#endif // _OXS_KEY
