/* FILE: mesh.cc                 -*-Mode: c++-*-
 *
 * Abstract Oxs_Mesh class, derives from Oxs_Ext and Oxs_Lock classes.
 *
 */

#include "mesh.h"
#include "meshvalue.h"
#include "vf.h"

/* End includes */


////////////////////////////////////////////////////////////////////////
// Oxs_Mesh (generic mesh) class

// Constructors
Oxs_Mesh::Oxs_Mesh
( const char* name,     // Child instance id
  Oxs_Director* newdtr  // App director
  ) : Oxs_Ext(name,newdtr)
{}

Oxs_Mesh::Oxs_Mesh
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr    // MIF block argument string
  ) : Oxs_Ext(name,newdtr,safe_interp,argstr)
{}


//////////////////////////////////////////////////
// Default file output using generic Oxs_Mesh interface.
// Returns 0 on success.
int Oxs_Mesh::WriteOvfIrregular
(const char* filename,  // Name of output file
 const char* title,     // Long filename or title
 const char* desc,      // Description to embed in output file
 const char* datatype,  // Either "binary" or "text"
 const char* precision, // For binary, "4" or "8";
 //  for text, a printf-style format
 const Oxs_MeshValue<ThreeVector>* vec,  // Vector array
 const Oxs_MeshValue<REAL8m>* scale,     // Optional scaling for vec
                    /// Set scale to NULL to use vec values directly.
 const ThreeVector* stephints  // If non-null, then the values are
                              /// written as stephints to the file.
 ) const
{
  // Check import validity
  enum DataType { BINARY, TEXT };
  DataType dt = BINARY;
  const char* filemode = "wb";
  if(strcmp("text",datatype)==0) {
    dt = TEXT;
    filemode = "w";
  } else if(strcmp("binary",datatype)!=0){
    return 1;  // Bad datatype
  }

  int datawidth = 0;
  string dataformat;
  if(dt == BINARY) {
    datawidth = atoi(precision);
    if(datawidth != 4 && datawidth != 8) {
      return 2; // Bad precision
    }
  } else {
    if(precision==NULL || precision[0]=='\0')
      precision="%# .17g";  // Default format
    string temp = string(precision) + string(" ");
    dataformat = temp + temp + temp + temp + temp +
      string(precision) + string("\n");
  }

  if(!vec->CheckMesh(this) || (scale!=NULL && scale->CheckMesh(this))) {
    return 3; // Import data not compatible with *this
  }

  // Open file
  Oxs_FilePointer file(filename,filemode);

  // Write header
  fprintf(file.fptr,"# OOMMF: irregular mesh v1.0\n");
  fprintf(file.fptr,"# Segment count: 1\n");
  fprintf(file.fptr,"# Begin: Segment\n");
  fprintf(file.fptr,"# Begin: Header\n");
  if(title==NULL || title[0]=='\0')
    fprintf(file.fptr,"# Title: %s\n",filename);
  else
    fprintf(file.fptr,"# Title: %s\n",title);
  if(desc!=NULL && desc[0]!='\0') {
    // Print out description, breaking at newlines as necessary.
    // Note: This block is optional
    const char *cptr1,*cptr2;
    cptr1=desc;
    while((cptr2=strchr(cptr1,'\n'))!=NULL) {
      fprintf(file.fptr,"# Desc: %.*s\n",(int)(cptr2-cptr1),cptr1);
      cptr1=cptr2+1;
    }
    if(*cptr1!='\0') fprintf(file.fptr,"# Desc: %s\n",cptr1);
  }
  fprintf(file.fptr,"# meshunit: m\n");
  fprintf(file.fptr,"# valueunit: A/m\n");
  fprintf(file.fptr,"# valuemultiplier: 1.0\n");

  Oxs_Box bbox;
  GetBoundingBox(bbox);
  fprintf(file.fptr,
	  "# xmin: %.16g\n# ymin: %.16g\n# zmin: %.16g\n"
	  "# xmax: %.16g\n# ymax: %.16g\n# zmax: %.16g\n",
	  bbox.GetMinX(),bbox.GetMinY(),bbox.GetMinZ(),
	  bbox.GetMaxX(),bbox.GetMaxY(),bbox.GetMaxZ());

  // Skip "boundary:" for now
  REAL8m minmag=0,maxmag=0;
  if(Size()>0) {
    REAL8m tempscale=1.0;
    maxmag=(*vec)[0].MagSq();
    if(scale!=NULL) {
      tempscale=(*scale)[0];
      maxmag*=tempscale*tempscale;
    }
    minmag=maxmag;
    for(UINT4m i=1;i<Size();i++) {
      REAL8m val=(*vec)[i].MagSq();
      if(scale!=NULL) {
	tempscale=(*scale)[i];
	val*=tempscale*tempscale;
      }
      if(val<minmag) minmag=val;
      if(val>maxmag) maxmag=val;
    }
    maxmag=sqrt(maxmag);
    minmag=sqrt(minmag);
    minmag*=0.999;  // Underestimate lower bound by 0.1% to protect against
    /// rounding errors.
  }

  fprintf(file.fptr,"# ValueRangeMaxMag: %.16g\n",maxmag);
  fprintf(file.fptr,"# ValueRangeMinMag: %.16g\n",minmag*1e-8);
  /// minmag is multiplied by 1e-8 to specify contrast range.

  fprintf(file.fptr,"# meshtype: irregular\n");
  fprintf(file.fptr,"# pointcount: %lu\n",
          static_cast<unsigned long>(Size()));

  // Write stephints, if requested
  if(stephints!=NULL) {
    fprintf(file.fptr,"# xstepsize: %.16g\n",stephints->x);
    fprintf(file.fptr,"# ystepsize: %.16g\n",stephints->y);
    fprintf(file.fptr,"# zstepsize: %.16g\n",stephints->z);
  }


  fprintf(file.fptr,"# End: Header\n");

  // Write data block
  if(dt == BINARY) {
    if(datawidth==4) {
      fprintf(file.fptr,"# Begin: Data Binary 4\n");
      REAL4 buf[6];
      buf[0]=1234567.;  // 4-Byte checkvalue
      Vf_OvfFileFormatSpecs::WriteBinary(file.fptr,buf,1);
      INT4m size=Size();
      for(INT4m i=0 ; i<size ; ++i) {
	ThreeVector loc;   Center(i,loc);
        buf[0] = REAL4(loc.x);
	buf[1] = REAL4(loc.y);
	buf[2] = REAL4(loc.z);
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  buf[3] = REAL4(v.x);
	  buf[4] = REAL4(v.y);
	  buf[5] = REAL4(v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  buf[3] = REAL4(tempscale*v.x);
	  buf[4] = REAL4(tempscale*v.y);
	  buf[5] = REAL4(tempscale*v.z);
	}
        Vf_OvfFileFormatSpecs::WriteBinary(file.fptr,buf,6);
	// Vf_OvfFileFormatSpecs::WriteBinary performs
	// byte-swapping as needed.
      }
      fprintf(file.fptr,"\n# End: Data Binary 4\n");
    } else {
      fprintf(file.fptr,"# Begin: Data Binary 8\n");
      REAL8 buf[6];
      buf[0]=123456789012345.;  // 8-Byte checkvalue
      Vf_OvfFileFormatSpecs::WriteBinary(file.fptr,buf,1);
      INT4m size=Size();
      for(INT4m i=0 ; i<size ; ++i) {
	ThreeVector loc;   Center(i,loc);
        buf[0] = REAL4(loc.x);
	buf[1] = REAL4(loc.y);
	buf[2] = REAL4(loc.z);
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  buf[3] = REAL4(v.x);
	  buf[4] = REAL4(v.y);
	  buf[5] = REAL4(v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  buf[3] = REAL4(tempscale*v.x);
	  buf[4] = REAL4(tempscale*v.y);
	  buf[5] = REAL4(tempscale*v.z);
	}
        Vf_OvfFileFormatSpecs::WriteBinary(file.fptr,buf,6);
	// Vf_OvfFileFormatSpecs::WriteBinary performs
	// byte-swapping as needed.
      }
      fprintf(file.fptr,"\n# End: Data Binary 8\n");
    }
  } else {
    fprintf(file.fptr,"# Begin: Data Text\n");
    INT4m size=Size();
    for(INT4m i=0 ; i<size ; ++i) {
	ThreeVector loc;   Center(i,loc);
	const ThreeVector& v = (*vec)[i];
	if(scale==NULL) {
	  fprintf(file.fptr,dataformat.c_str(),
		  loc.x,loc.y,loc.z,v.x,v.y,v.z);
	} else {
	  REAL8m tempscale=(*scale)[i];
	  fprintf(file.fptr,dataformat.c_str(),
		  loc.x,loc.y,loc.z,
		  tempscale*v.x,tempscale*v.y,tempscale*v.z);
	}
    }
    fprintf(file.fptr,"# End: Data Text\n");
  }

  fprintf(file.fptr,"# End: Segment\n");
  return 0;
}

//////////////////////////////////////////////////
// Default volume summing routines using generic Oxs_Mesh interface.
REAL8m
Oxs_Mesh::VolumeSum(const Oxs_MeshValue<REAL8m>* scalar) const
{
  if(!scalar->CheckMesh(this)) {
    string msg="Incompatible scalar array import to"
      " VolumeSum(const Oxs_MeshValue<REAL8m>*) in object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  UINT4m size=Size();
  REAL8m sum=0.0;
  for(UINT4m i=0;i<size;i++) {
    sum += (*scalar)[i] * Volume(i);
  }
  return sum;
}

ThreeVector
Oxs_Mesh::VolumeSum(const Oxs_MeshValue<ThreeVector>* vec,
		    const Oxs_MeshValue<REAL8m>* scale) const
{
  if(!vec->CheckMesh(this)) {
    string msg="Incompatible vector array import to"
      " VolumeSum(const Oxs_MeshValue<ThreeVector>*) in object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  if(scale!=NULL && !scale->CheckMesh(this)) {
    string msg="Incompatible scaling array import to"
      " VolumeSum(const Oxs_MeshValue<ThreeVector>*) in object "
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  UINT4m size=Size();
  ThreeVector sum(0.,0.,0.);
  if(scale==NULL) {
    for(UINT4m i=0;i<size;i++) {
      sum +=  ( Volume(i) * (*vec)[i] );
    }
  } else {
    for(UINT4m i=0;i<size;i++) {
      sum +=  ( (Volume(i) * (*scale)[i]) * (*vec)[i] );
    }
  }
  return sum;
}
