/* FILE: oxscmd.cc                 -*-Mode: c++-*-
 *
 * Tcl interface to C/C++ code in oxs.
 *
 * Each routine in the file is registered with the Tcl interpreter
 * by OxsRegisterInterfaceCommands() at the bottom of this file.
 * OxsRegisterInterfaceCommands() is called by Oxs_Init().
 * (Note: OxsRegisterInterfaceCommands() is not in any header file
 * because it should only be called by Oxs_Init().)
 *
 */

#include "director.h"
#include "ext.h"
#include "oc.h"
#include "energy.h"	// Needed to make MSVC++ 5 happy

/* End includes */     

/*
 *----------------------------------------------------------------------
 *
 * Program globals
 *
 *----------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbInit --
 *
 *      Micromagnetic problem initialization
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

extern "C" int Oxs_ProbInit(ClientData cd,Tcl_Interp *interp,
			    int argc,char **argv)
{
  if (argc != 2) {
    Tcl_AppendResult(interp, "wrong # of args: should be",
		     " \"Oxs_ProbInit mif_object\"",
		     (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  return director->ProbInit(argv[1]);
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbReset --
 *
 *      Resets micromagnetic problem
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
extern "C" int Oxs_ProbReset(ClientData cd,Tcl_Interp *interp,
			     int argc,char**)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be just",
		     " \"Oxs_ProbReset\"",
		     (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  return director->ProbReset();
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ProbRelease --
 *
 *      Releases micromagnetic problem statement.  Depending on OS,
 *      this may free some memory, but this routine is intended
 *      primarily as a debugging aid.
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
extern "C" int Oxs_ProbRelease(ClientData cd,Tcl_Interp *interp,
			     int argc,char**)
{
  if (argc != 1) {
    Tcl_AppendResult(interp, "wrong # of args: should be just",
		     " \"Oxs_ProbRelease\"",
		     (char *) NULL);
    return TCL_ERROR;
  }
  Oxs_Director* director = (Oxs_Director *)cd;
  return director->ProbRelease();
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListRegisteredExtClasses --
 *
 *      Lists Ext child classes that are loaded and registered
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
extern "C" int Oxs_ListRegisteredExtClasses(ClientData,
					    Tcl_Interp *interp,
					    int,char**)
{
  string result=Oxs_Ext::ListRegisteredChildClasses();
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,result.c_str(),(char *)NULL);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListExtObjects --
 *
 *      Lists Ext objects that have been created
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
extern "C" int Oxs_ListExtObjects(ClientData cd,Tcl_Interp *interp,
				  int,char**)
{
  Oxs_Director* director = (Oxs_Director *)cd;
  string result=director->ListExtObjects();
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,result.c_str(),(char *)NULL);
  return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Oxs_ListEnergyObjects --
 *
 *      Lists Energy objects that have been created
 *
 * Results:
 *      A standard Tcl result.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
extern "C" int Oxs_ListEnergyObjects(ClientData cd,Tcl_Interp *interp,
				     int,char**)
{
  Oxs_Director* director = (Oxs_Director *)cd;
  string result=director->ListEnergyObjects();
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,result.c_str(),(char *)NULL);
  return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * OxsDeleteDirector --
 *
 *      Deletes Oxs_Director instance that has been associated with
 *      a Tcl_Interp.  This routine is called by the interpreter
 *      as part of its destruction.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
Tcl_InterpDeleteProc OxsDeleteDirector;

void OxsDeleteDirector(ClientData clientdata,Tcl_Interp *interp)
{
  Oxs_Director *dir = (Oxs_Director *)clientdata;
  if(dir->VerifyInterp(interp)) {
    delete dir;
  }
  // Else, throw exception?
}

/*
 *----------------------------------------------------------------------
 *
 * OxsRegisterInterfaceCommands --
 *
 *      Registers each of the above routines with the Tcl interpreter
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
void OxsRegisterInterfaceCommands(Oxs_Director* director,
				  Tcl_Interp *interp)
{
#define REGCMD(foo) \
  Tcl_CreateCommand(interp,Oc_AutoBuf(#foo),foo,director,NULL)

  REGCMD(Oxs_ProbInit);
  REGCMD(Oxs_ProbReset);
  REGCMD(Oxs_ProbRelease);
  REGCMD(Oxs_ListRegisteredExtClasses);
  REGCMD(Oxs_ListEnergyObjects);
  REGCMD(Oxs_ListExtObjects);

#undef REGCMD
}

