/* FILE: if.cc                      -*-Mode: c++-*-
 *
 *	The OOMMF Image Formats extension.
 *
 *	This extension supplies classes which can manipulate
 * various image formats.  When the extension is loaded into
 * an interpreter in which Tk has been loaded, it registers
 * corresponding photo image formats with Tk.
 * 
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2000/08/11 23:24:09 $
 * Last modified by: $Author: dgp $
 */

#include <stdio.h>
#include <stdlib.h>

#define USE_OLD_IMAGE
#include "oc.h"
#include "if.h"

/* End includes */     // Optional directive to pimake

struct RGBQuad {
  BYTE Red;
  BYTE Green;
  BYTE Blue;
  BYTE Reserved;
  int MSFill32(Tcl_Channel chan);  // Fills structure from chan
  /// of MS RGBQuad's, returning the number of bytes read (4),
  /// or 0 on error.
  inline int MSFill24(const char* carr); // Analogous to above,
  /// but takes input from a 3-byte long buffer.
};

int RGBQuad::MSFill32(Tcl_Channel chan)
{ // Note: MS RGBQuad's are ordered "Blue Green Red Reserved"
  char carr[4];
  int count=Tcl_Read(chan,carr,4);
  if(count!=4) return 0;
  Blue=carr[0];    Green=carr[1];
  Red=carr[2];     Reserved=carr[3];
  return count;
}

int RGBQuad::MSFill24(const char* carr)
{ // Note: MS RGBQuad's are ordered "Blue Green Red Reserved"
  Blue=carr[0];    Green=carr[1];
  Red=carr[2];     Reserved=0;
  return 3;
}

class MSBitmap;  // Forward declaration for typedef
typedef int (MSBitmap::*BmpConvert)(const char* read_buf,
				    RGBQuad* &pix,
				    UINT4 startcol,
				    UINT4 stopcol);

class MSBitmap {
private:
  BYTE  Type[2];
  UINT4 FileSize;
  UINT2 Reserved1;
  UINT2 Reserved2;
  UINT4 OffBits;
  UINT4 BmiSize;
  UINT4 Width;
  UINT4 Height;
  UINT2 Planes;
  UINT2 BitCount;
  UINT4 Compression;
  UINT4 SizeImage;
  UINT4 XPelsPerMeter;
  UINT4 YPelsPerMeter;
  UINT4 ClrUsed;
  UINT4 ClrImportant;
  RGBQuad* Palette;

  int AllocPalette(int size); // Returns 1 on success
  void FreePalette();

  int FillHeader(Tcl_Channel chan,BOOL fillpalette);
  /// If 'fillpallette' is true, then fills all data members;
  /// If 'fillpalette' is 0, then fill everything *except*
  /// the palette and related members.
  ///   Returns 1 if successful, 0 if the input does not appear
  /// to be a valid MS bitmap.  Adjusts for alignment and byte
  /// ordering as needed.

  // Support variables, computed from the above
  // Calculate number of bytes in per row in file; by spec,
  // must be divisible by 4.
  UINT4 FileRowSize;
  UINT4 PaletteSize;

  // FillPhoto() bitmap->rgb conversion routines.
  int  Bmp1toRgbq(const char* read_buf,RGBQuad* &pix,
		  UINT4 startcol,UINT4 stopcol);        //  1 bit/pixel
  int  Bmp4toRgbq(const char* read_buf,RGBQuad* &pix,
		  UINT4 startcol,UINT4 stopcol);        //  4 bits/pixel
  int  Bmp8toRgbq(const char* read_buf,RGBQuad* &pix,
		  UINT4 startcol,UINT4 stopcol);        //  8 bits/pixel
  int  Bmp24toRgbq(const char* read_buf,RGBQuad* &pix,
		  UINT4 startcol,UINT4 stopcol);        // 24 bits/pixel
  BmpConvert DataConvert;

public:
  MSBitmap();
  ~MSBitmap();
  int ReadCheck(Tcl_Channel chan,int& imagewidth,int& imageheight);
  /// Returns 1 if the file on chan looks like a Microsoft BMP
  /// file that we know how to read.

  int FillPhoto(Tcl_Interp*  interp,Tcl_Channel  chan,
	   char* fileName,Tk_PhotoHandle imageHandle,
	   int destX,int destY,int width,int height,int srcX,int srcY);
  /// Fills in imageHandle as requested, one row at a time.

};

int MSBitmap::AllocPalette(int size)
{
  FreePalette();
  if((Palette=new RGBQuad[size])!=NULL) {
    PaletteSize=UINT4(size);
    return 1;
  }
  return 0;
}

void MSBitmap::FreePalette()
{
  if(Palette!=NULL) { delete[] Palette; Palette=NULL; }
  PaletteSize=0;
}

MSBitmap::~MSBitmap() { FreePalette(); }

MSBitmap::MSBitmap() :
  FileSize(0),Reserved1(0),Reserved2(0),OffBits(0),
  BmiSize(0),Width(0),Height(0),Planes(0),BitCount(0),Compression(0),
  SizeImage(0),XPelsPerMeter(0),YPelsPerMeter(0),
  ClrUsed(0),ClrImportant(0),Palette(NULL),
  FileRowSize(0),PaletteSize(0),DataConvert(BmpConvert(NULL))
{
  Type[0]=Type[1]='\0';
}

int MSBitmap::FillHeader(Tcl_Channel chan,BOOL fillpalette)
{ // Reads members one at a time to avoid differing packing
  // alignment restrictions on various machines.  Returns 1
  // on success.

  // Read and check signature
  Tcl_Read(chan,(char *)Type,2*sizeof(BYTE));
  if(Type[0]!='B' || Type[1]!='M') return 0;

  // Read rest of header
  Tcl_Read(chan,(char *)&FileSize,sizeof(UINT4));
  Tcl_Read(chan,(char *)&Reserved1,sizeof(UINT2));
  Tcl_Read(chan,(char *)&Reserved2,sizeof(UINT2));
  Tcl_Read(chan,(char *)&OffBits,sizeof(UINT4));

  // MS Bitmap format is little-endian ordered;
  // Adjust byte ordering, if necessary.
#if BYTEORDER != 4321
#if BYTEORDER != 1234
#error "Unsupported byte-order format"
#endif
  Oc_Flip4(&FileSize);
  Oc_Flip2(&Reserved1); Oc_Flip2(&Reserved2);
  Oc_Flip4(&OffBits);
#endif

  // Check file size.  Is there no Tcl C interface to the 'file size'
  // proc???
  int bmistart=Tcl_Tell(chan);
  int real_size=Tcl_Seek(chan,0,SEEK_END);
  Tcl_Seek(chan,bmistart,SEEK_SET);
  if(real_size!=-1 && (UINT4)real_size!=FileSize) return 0;

  // Otherwise, read in rest of header
  Tcl_Read(chan,(char *)&BmiSize,sizeof(UINT4));
  Tcl_Read(chan,(char *)&Width,sizeof(UINT4));
  Tcl_Read(chan,(char *)&Height,sizeof(UINT4));
  Tcl_Read(chan,(char *)&Planes,sizeof(UINT2));
  Tcl_Read(chan,(char *)&BitCount,sizeof(UINT2));
  Tcl_Read(chan,(char *)&Compression,sizeof(UINT4));
  Tcl_Read(chan,(char *)&SizeImage,sizeof(UINT4));
  Tcl_Read(chan,(char *)&XPelsPerMeter,sizeof(UINT4));
  Tcl_Read(chan,(char *)&YPelsPerMeter,sizeof(UINT4));
  Tcl_Read(chan,(char *)&ClrUsed,sizeof(UINT4));
  Tcl_Read(chan,(char *)&ClrImportant,sizeof(UINT4));

  // MS Bitmap format is little-endian ordered;
  // Adjust byte ordering, if necessary.
#if BYTEORDER != 4321
#if BYTEORDER != 1234
#error "Unsupported byte-order format"
#endif
  Oc_Flip4(&BmiSize);
  Oc_Flip4(&Width);          Oc_Flip4(&Height);
  Oc_Flip2(&Planes);         Oc_Flip2(&BitCount);
  Oc_Flip4(&Compression);    Oc_Flip4(&SizeImage);
  Oc_Flip4(&XPelsPerMeter);  Oc_Flip4(&YPelsPerMeter);
  Oc_Flip4(&ClrUsed);        Oc_Flip4(&ClrImportant);
#endif

  FileRowSize=4*((Width*BitCount+31)/32); // Rows lie on a 4-byte bdry.

  switch(BitCount)
    {
    case 1:   DataConvert = &MSBitmap::Bmp1toRgbq;        break;
    case 4:   DataConvert = &MSBitmap::Bmp4toRgbq;        break;
    case 8:   DataConvert = &MSBitmap::Bmp8toRgbq;        break;
    case 24:  DataConvert = &MSBitmap::Bmp24toRgbq;       break;
    default:  DataConvert = BmpConvert(NULL);  break;
    }

  // If requested, read palette info from file
  FreePalette();
  if(fillpalette) {
    // Position at end of BITMAPINFOHEADER
    Tcl_Seek(chan,bmistart+BmiSize,SEEK_SET);

    int palsize=ClrUsed;
    if(palsize==0) {
      switch(BitCount)
	{
	case 1:  palsize=2;    break;
	case 4:  palsize=16;   break;
	case 8:  palsize=256;  break;
	case 24: palsize=0;    break;  // True color
	default: return 0;     // Invalid BitCount value
	}
    }
    if(palsize>0 &&
       AllocPalette(palsize)!=0) {
      for(int i=0;i<palsize;i++)
	Palette[i].MSFill32(chan);
    }
  }

  return 1;
}

int
MSBitmap::Bmp1toRgbq(const char* read_buf,RGBQuad* &pix,
		     UINT4 startcol,UINT4 stopcol)
{ // 1 bit per pixel, with palette
  UINT4 jstart=8*(startcol/8);
  UINT4 jstop=8*((stopcol+7)/8);
  UINT4 j=jstart;
  RGBQuad offbit=Palette[0];
  RGBQuad onbit=Palette[1];
  while(j<jstop) {
    unsigned int datum=read_buf[j/8];
    for(unsigned int mask=0x80; mask!=0 ; mask>>=1) {
      // Note bit-order!
      if(j<stopcol && j>=startcol) {
	if((datum & mask)==0) pix[j-startcol]=offbit;
	else                  pix[j-startcol]=onbit;
      }
      j++;
    }
  }
  return 0;
}

int
MSBitmap::Bmp4toRgbq(const char* read_buf,RGBQuad* &pix,
		     UINT4 startcol,UINT4 stopcol)
{ // 4 bits per pixel, with palette
  UINT4 jstart=2*(startcol/2);
  UINT4 jstop=2*((stopcol+1)/2);
  UINT4 j=jstart;
  while(j<jstop) {
    unsigned int datum=read_buf[j/2];
    UINT4 index1= datum & 0x0F;        // Note bit-order!
    UINT4 index0= (datum & 0xF0)>>4;
    if(startcol<=j) {
      if(index0>=PaletteSize) return 1;
      pix[j-startcol]=Palette[index0];
    }
    j++;
    if(j<stopcol) {
      if(index1>=PaletteSize) return 1;
      pix[j-startcol]=Palette[index1];
    }
    j++;
  }
  return 0;
}

int
MSBitmap::Bmp8toRgbq(const char* read_buf,RGBQuad* &pix,
		     UINT4 startcol,UINT4 stopcol)
{ // 8 bits per pixel, with palette
  for(UINT4 j=startcol;j<stopcol;j++) {
    // Copy values from palette
    UINT4 index=read_buf[j];
    if(index>=PaletteSize) return 1;
    pix[j-startcol]=Palette[index];
  }
  return 0;
}

int
MSBitmap::Bmp24toRgbq(const char* read_buf,RGBQuad* &pix,
		      UINT4 startcol,UINT4 stopcol)
{ // True color input file; 24 bits per pixel, no palette
  const int data_width=3;
  for(UINT4 j=startcol;j<stopcol;j++) {
    pix[j-startcol].MSFill24(read_buf+j*data_width);
  }
  return 0;
}

int MSBitmap::ReadCheck(Tcl_Channel chan,
			int& imagewidth,int& imageheight)
{ // Returns 1 if the file on chan looks like a Microsoft BMP file that
  // we know how to read.
  if(FillHeader(chan,0)==0) return 0; // Invalid header

  // Otherwise, looks like an MS bitmap file.  Check that it is a
  // subformat we support.
  if(Planes!=1         ||      // Only known value
     Compression!=0    ||      // No compression
     DataConvert==BmpConvert(NULL)) { // Bits per pixel check
    return 0;
  }

  imagewidth=Width;
  imageheight=Height;

  return 1;
}

int
MSBitmap::FillPhoto(Tcl_Interp*  interp,Tcl_Channel  chan,
		    char* fileName,Tk_PhotoHandle imageHandle,
		    int destX,int destY,int reqwidth,int reqheight,
		    int srcX,int srcY)
{ // Fills in imageHandle as requested, one row at a time.
  // Returns TCL_OK on success, TCL_ERROR on failure.

  if(FillHeader(chan,1) == 0) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,"Header in file ",fileName,
                     " not recognized as a Microsoft .bmp header",
                     (char *)NULL);
    return TCL_ERROR;
  }

  // Looks like an MS bitmap file.  Check that it is a subformat
  // we support.
  if(Planes!=1         ||      // Only known value
     Compression!=0    ||      // No compression
     DataConvert==BmpConvert(NULL)) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,"Sub-format of Microsoft .bmp file ",
                     fileName," is not currently supported.",
                     (char *)NULL);
    return TCL_ERROR;
  }

  // Deduce working control parameters
  int stoprow  = Height - srcY;
  int startrow = stoprow - reqheight;
  if(startrow<0) { reqheight += startrow; startrow=0; }
  if(UINT4(stoprow)>Height) {
    reqheight -= stoprow-Height;   // Safety
    stoprow = Height;
  }
  int startcol = srcX;
  int stopcol = startcol+reqwidth;
  if(startcol<0) { reqwidth += startcol; startcol=0; }
  if(UINT4(stopcol)>Width) {
    reqwidth -= stopcol-Width;
    stopcol = Width;
  }
  if(startrow>=stoprow || startcol>=stopcol) {
    return TCL_OK; // Nothing to do
  }

  // 
  char *read_buf=new char[FileRowSize];
  RGBQuad *pix=new RGBQuad[reqwidth];

  Tk_PhotoImageBlock pib;
  pib.pixelPtr=(unsigned char *)pix;
  pib.width=reqwidth;
  pib.height=1;
  pib.pitch=int(sizeof(RGBQuad)*Width);
  pib.pixelSize=sizeof(RGBQuad);
  pib.offset[0]=int(((unsigned char*)&pix[0].Red)   - (unsigned char*)pix);
  pib.offset[1]=int(((unsigned char*)&pix[0].Green) - (unsigned char*)pix);
  pib.offset[2]=int(((unsigned char*)&pix[0].Blue)  - (unsigned char*)pix);

  // Skip unrequested leading rows
  Tcl_Seek(chan,OffBits+startrow*FileRowSize,SEEK_SET);

  for(UINT4 i= UINT4(startrow);i<UINT4(stoprow);i++) {
    if(Tcl_Read(chan,read_buf,FileRowSize)==-1) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,"Input error during read of"
                       " Microsoft .bmp file ",fileName,
                       " bitmap.",(char *)NULL);
      return TCL_ERROR;
    }
    switch((this->*DataConvert)(read_buf,pix,UINT4(startcol),UINT4(stopcol)))
      {
      case 0:  break;
      default:
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp,"Input error during read of"
			 " Microsoft .bmp file ",fileName,
			 " bitmap: Illegal palette index",
			 (char *)NULL);
	return TCL_ERROR;
      }
    Tk_PhotoPutBlock(imageHandle,&pib,destX,destY+stoprow-i-1,
		     reqwidth,1);
  }
  delete[] pix;
  delete[] read_buf;

  return TCL_OK;
}

// The C interface for extending the photo image formats changed
// incompatibly between versions 4 and 8 of Tk.
// May want to code a workaround eventually.  For now, the bmp
// photo image format is simply not available in pre-8.0 interpreters.
#if (TK_MAJOR_VERSION >= 8)
extern "C" {
  static Tk_ImageFileMatchProc bmpFileMatchProc;
  static Tk_ImageFileReadProc bmpFileReadProc;
}


static char bmpnamestr[] = "bmp";

int
bmpFileMatchProc(Tcl_Channel chan,
                 char*        /* fileName */,
                 char*        /* formatString */,
                 int         *widthPtr,
                 int         *heightPtr)
{
  MSBitmap msb;
  return msb.ReadCheck(chan,*widthPtr,*heightPtr);
}

int
bmpFileReadProc(Tcl_Interp*  interp,
                Tcl_Channel  chan,
                char*        fileName,
                char*        /* formatString */,
                Tk_PhotoHandle imageHandle,
                int destX,int destY,
                int width,int height,
                int srcX,int srcY)
{
  MSBitmap msb;
  return msb.FillPhoto(interp,chan,fileName,imageHandle,
		       destX,destY,width,height,srcX,srcY);
}


static Tk_PhotoImageFormat bmpformat = 
{
  bmpnamestr,             // Format name
  bmpFileMatchProc,       // File format identifier routine
  NULL,                   // String format id routine
  bmpFileReadProc,        // File read routine
  NULL,                   // String read routine
  NULL,                   // File write routine
  NULL,                    // String write routine
  NULL,			  // NULL nextPtr to be overwritten by Tk
};
#endif

int 
If_Init(Tcl_Interp *interp)
{
#define RETURN_TCL_ERROR                                               \
    Tcl_AddErrorInfo(interp, ab("\n    (in If_Init())"));              \
    return TCL_ERROR

    Oc_AutoBuf ab, ab0, ab1, ab2, ab3;

    if (Tcl_PkgPresent(interp, ab("Oc"), ab1("1.1"), 0) == NULL) {
        Tcl_AppendResult(interp, ab("\n\t(If "IF_VERSION" needs Oc 1.1)"),
                NULL);
        RETURN_TCL_ERROR;
    }

// See note above.
#if (TK_MAJOR_VERSION >= 8)
    Tk_CreatePhotoImageFormat(&bmpformat);
#endif

    if (Tcl_PkgProvide(interp, ab("If"), ab1(IF_VERSION)) != TCL_OK) {
        RETURN_TCL_ERROR;
    }

/*
 * Currently there is no Tcl part to the If extension
 *
    if (Oc_InitScript(interp, "If", IF_VERSION) != TCL_OK) {
        RETURN_TCL_ERROR;
    }
 *
 */

    return TCL_OK;

#undef RETURN_TCL_ERROR
}

